"""
Gantry's public API.
"""

import os
from contextlib import ExitStack
from pathlib import Path
from typing import Any, Literal

from beaker import Beaker

from . import beaker_utils
from .callbacks import *
from .git_utils import GitRepoState
from .launch import follow_workload, launch_experiment
from .recipe import Recipe

__all__ = [
    "Recipe",
    "GitRepoState",
    "Callback",
    "SlackCallback",
    "launch_experiment",
    "follow_workload",
    "update_workload_description",
    "write_metrics",
]


_ORIGINAL_WORKLOAD_DESCRIPTIONS: dict[str, str] = {}


def update_workload_description(
    description: str,
    strategy: Literal["append", "prepend", "replace"] = "replace",
    beaker_token: str | None = None,
    client: Beaker | None = None,
) -> str:
    """
    Update the description of the Gantry workload that this process is running in.

    :param description: The description to set or add, depending on the ``strategy``.
    :param strategy: One of "append", "prepend", or "replace" to indicate how the new description
        should be combined with the original description. Defaults to "replace".
    :param beaker_token: An optional Beaker API token to use. If not provided, the
        ``BEAKER_TOKEN`` environment variable will be used if set, or a Beaker config file.
        Alternatively you can provide an existing :class:`~beaker.Beaker` client via the
        ``client`` parameter.
    :param client: An optional existing :class:`~beaker.Beaker` client to use. If not provided,
        a new client will be created using the provided ``beaker_token`` or environment/config.
    """
    global _ORIGINAL_WORKLOAD_DESCRIPTIONS

    if (workload_id := os.environ.get("BEAKER_WORKLOAD_ID")) is None:
        raise RuntimeError(
            "'update_workload_description' can only be called from within a running workload"
        )

    with ExitStack() as stack:
        if client is None:
            beaker: Beaker = stack.enter_context(
                beaker_utils.init_client(
                    ensure_workspace=False, beaker_token=beaker_token, check_for_upgrades=False
                )
            )
        else:
            beaker = client

        workload = beaker.workload.get(workload_id)
        if workload_id not in _ORIGINAL_WORKLOAD_DESCRIPTIONS:
            _ORIGINAL_WORKLOAD_DESCRIPTIONS[workload_id] = (
                workload.experiment.description or ""
            ).strip()

        og_description = _ORIGINAL_WORKLOAD_DESCRIPTIONS[workload_id]

        if strategy == "append":
            description = og_description + " " + description
        elif strategy == "prepend":
            description = description + " " + og_description
        elif strategy != "replace":
            raise ValueError(
                f"'strategy' must be one of 'append', 'prepend', or 'replace', but got '{strategy}'."
            )

        description = description.strip()
        beaker.workload.update(workload, description=description)

    return description


def write_metrics(metrics: dict[str, Any]):
    """
    Write result metrics for the Gantry workload that this process is running in.

    :param metrics: A JSON-serializable dictionary of metrics to write.
    """
    import json

    if os.environ.get("BEAKER_WORKLOAD_ID") is None:
        raise RuntimeError("'write_metrics' can only be called from within a running workload")
    if (results_dir := os.environ.get("RESULTS_DIR")) is None:
        raise RuntimeError("Results directory not set! Can't write metrics.")
    metrics_path = Path(results_dir) / "metrics.json"
    metrics_path.parent.mkdir(exist_ok=True, parents=True)
    with metrics_path.open("w") as f:
        json.dump(metrics, f)



from typing import Sequence

import click
from beaker import BeakerWorkload
from beaker.exceptions import BeakerWorkloadNotFound
from rich import prompt

from .. import beaker_utils, utils
from ..config import get_global_config
from ..exceptions import ConfigurationError, NotFoundError
from .main import CLICK_COMMAND_DEFAULTS, main


@main.command(**CLICK_COMMAND_DEFAULTS)
@click.argument("workload", nargs=-1, type=str)
@click.option(
    "-l", "--latest", is_flag=True, help="""Stop your latest experiment (non-session) workload."""
)
@click.option(
    "-w",
    "--workspace",
    type=str,
    help=f"""The Beaker workspace to pull experiments from. {get_global_config().get_help_string_for_default('workspace')}""",
)
@click.option("--dry-run", is_flag=True, help="Do a dry-run without stopping any experiments.")
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    help="""Skip all confirmation prompts.""",
)
def stop(
    workload: Sequence[str],
    latest: bool = False,
    workspace: str | None = None,
    dry_run: bool = False,
    yes: bool = False,
):
    """
    Stop a running workload.
    """
    beaker = beaker_utils.init_client(ensure_workspace=False)
    workloads: list[BeakerWorkload] = []
    if workload:
        if latest or workspace is not None:
            raise ConfigurationError(
                "[WORKLOAD] args are mutually exclusive with -w/--workspace and -l/--latest"
            )
        for workload_name in workload:
            try:
                workloads.append(beaker.workload.get(workload_name))
            except BeakerWorkloadNotFound:
                raise NotFoundError(f"Workload '{workload_name}' not found")
    else:
        if not latest:
            raise ConfigurationError(
                "A filter such as -l/--latest is required when no [WORKLOAD] is specified"
            )

        wl = beaker_utils.get_latest_workload(
            beaker, workspace_name=workspace or get_global_config().workspace, running=True
        )
        if wl is None:
            utils.print_stderr("[yellow]No running workloads to stop[/]")
        else:
            workloads.append(wl)

    for wl in workloads:
        if dry_run:
            utils.print_stdout(f"[b yellow]Dry run:[/] would stop [b cyan]{wl.experiment.name}[/]")
        else:
            if not yes and not prompt.Confirm.ask(
                f"Stop experiment [b cyan]{wl.experiment.name}[/] at [blue u]{beaker.workload.url(wl)}[/]?"
            ):
                utils.print_stdout("[yellow]Skipping experiment...[/]")
                continue

            try:
                beaker.workload.cancel(wl)
            except (BeakerWorkloadNotFound,):
                pass
            utils.print_stdout(
                f"[b green]\N{check mark}[/] [b cyan]{wl.experiment.name}[/] at [blue u]{beaker.workload.url(wl)}[/] stopped"
            )



import dataclasses
from dataclasses import dataclass
from typing import Any, Literal, Sequence

from beaker import Beaker, BeakerWorkload

from . import constants, utils
from .aliases import PathOrStr
from .callbacks import Callback
from .exceptions import *
from .git_utils import GitRepoState
from .launch import launch_experiment


@dataclass
class Recipe:
    """
    A recipe defines how Gantry creates a Beaker workload and can be used to programmatically
    launch Gantry runs from Python as opposed to from the command-line.
    """

    # Workload settings
    args: Sequence[str]
    name: str | None = None
    description: str | None = None
    workspace: str | None = None
    budget: str | None = None
    group_names: Sequence[str] | None = None

    # Launch settings.
    allow_dirty: bool = False
    yes: bool | None = None
    save_spec: PathOrStr | None = None

    # Callbacks.
    callbacks: Sequence[Callback] | None = None

    # Constraints.
    clusters: Sequence[str] | None = None
    gpu_types: Sequence[str] | None = None
    interconnect: Literal["ib", "tcpxo"] | None = None
    tags: Sequence[str] | None = None
    hostnames: Sequence[str] | None = None

    # Resources.
    cpus: float | None = None
    gpus: int | None = None
    memory: str | None = None
    shared_memory: str | None = None

    # Inputs.
    beaker_image: str | None = None
    docker_image: str | None = None
    datasets: Sequence[str] | None = None
    env_vars: Sequence[str | tuple[str, str]] | None = None
    env_secrets: Sequence[str | tuple[str, str]] | None = None
    dataset_secrets: Sequence[str | tuple[str, str]] | None = None
    mounts: Sequence[str | tuple[str, str]] | None = None
    weka: Sequence[str | tuple[str, str]] | None = None
    uploads: Sequence[str | tuple[str, str]] | None = None
    ref: str | None = None
    branch: str | None = None
    remote: str | None = None
    git_repo: GitRepoState | None = None
    gh_token_secret: str = constants.GITHUB_TOKEN_SECRET
    aws_config_secret: str | None = None
    aws_credentials_secret: str | None = None
    google_credentials_secret: str | None = None

    # Outputs.
    results: str = constants.RESULTS_DIR

    # Task settings.
    task_name: str = "main"
    priority: str | None = None
    task_timeout: str | None = None
    preemptible: bool | None = None
    retries: int | None = None

    # Multi-node config.
    replicas: int | None = None
    leader_selection: bool | None = None
    host_networking: bool | None = None
    propagate_failure: bool | None = None
    propagate_preemption: bool | None = None
    synchronized_start_timeout: str | None = None
    skip_tcpxo_setup: bool = dataclasses.field(default=False, repr=False)  # deprecated
    skip_nccl_setup: bool = False

    # Runtime.
    runtime_dir: str = constants.RUNTIME_DIR
    exec_method: Literal["exec", "bash"] = "exec"
    torchrun: bool = False

    # Setup hooks.
    pre_setup: str | None = None
    post_setup: str | None = None

    # Python settings.
    python_manager: Literal["uv", "conda"] | None = None
    default_python_version: str = utils.get_local_python_version()
    system_python: bool = False
    install: str | None = None
    no_python: bool = False

    # Python UV settings.
    uv_venv: str | None = None
    uv_extras: Sequence[str] | None = None
    uv_all_extras: bool | None = None
    uv_torch_backend: str | None = None
    uv_cache_dir: str | None = None

    # Python Conda settings.
    conda_file: PathOrStr | None = None
    conda_env: str | None = None

    @classmethod
    def multi_node_torchrun(
        cls,
        cmd: Sequence[str],
        gpus_per_node: int,
        num_nodes: int,
        shared_memory: str | None = "10GiB",
        **kwargs,
    ) -> "Recipe":
        """
        Create a multi-node recipe using torchrun.
        """
        return cls(
            args=cmd,
            gpus=gpus_per_node,
            replicas=num_nodes,
            shared_memory=shared_memory,
            torchrun=True,
            **kwargs,
        )

    def _get_launch_args(self) -> Sequence[str]:
        if isinstance(self.args, str):
            raise ConfigurationError("args must be a sequence of strings, not a single string")
        return self.args

    def _get_launch_kwargs(self) -> dict[str, Any]:
        kwargs = dataclasses.asdict(self)
        kwargs["callbacks"] = self.callbacks
        kwargs["git_repo"] = self.git_repo
        kwargs.pop("args")
        return kwargs

    def dry_run(self, client: Beaker | None = None) -> None:
        """
        Do a dry-run to validate options.
        """
        launch_experiment(
            self._get_launch_args(),
            **self._get_launch_kwargs(),
            client=client,
            dry_run=True,
        )

    def launch(
        self,
        show_logs: bool | None = None,
        timeout: int | None = None,
        start_timeout: int | None = None,
        inactive_timeout: int | None = None,
        inactive_soft_timeout: int | None = None,
        auto_cancel: bool = False,
        client: Beaker | None = None,
    ) -> BeakerWorkload:
        """
        Launch an experiment on Beaker. Same as the ``gantry run`` command.

        :returns: The Beaker workload.
        """
        workload = launch_experiment(
            self._get_launch_args(),
            **self._get_launch_kwargs(),
            show_logs=show_logs,
            timeout=timeout,
            start_timeout=start_timeout,
            inactive_timeout=inactive_timeout,
            inactive_soft_timeout=inactive_soft_timeout,
            auto_cancel=auto_cancel,
            client=client,
        )
        assert workload is not None
        return workload

    def with_replicas(
        self,
        replicas: int,
        leader_selection: bool = True,
        host_networking: bool = True,
        propagate_failure: bool = True,
        propagate_preemption: bool = True,
        synchronized_start_timeout: str = "5m",
        skip_nccl_setup: bool = False,
    ) -> "Recipe":
        """
        Add replicas to the recipe.
        """
        if replicas < 2:
            raise ConfigurationError("replicas must be at least 2")
        return dataclasses.replace(
            self,
            replicas=replicas,
            leader_selection=leader_selection,
            host_networking=host_networking,
            propagate_failure=propagate_failure,
            propagate_preemption=propagate_preemption,
            synchronized_start_timeout=synchronized_start_timeout,
            skip_nccl_setup=skip_nccl_setup,
        )



import dataclasses
from dataclasses import dataclass
from typing import Any, Literal

import tomli
from dataclass_extensions import decode

from .exceptions import ConfigurationError


@dataclass
class GantryConfig:
    workspace: str | None = None
    gh_token_secret: str | None = None
    budget: str | None = None
    log_level: Literal["debug", "info", "warning", "error"] | None = None
    quiet: bool | None = None
    callback_modules: list[str] | None = dataclasses.field(default=None)
    path: str | None = dataclasses.field(repr=False, default=None)

    @classmethod
    def load(cls) -> "GantryConfig":
        """
        Load configuration from a ``pyproject.toml`` under the ``[tool.gantry]`` section.
        """
        path = "pyproject.toml"
        try:
            with open(path, "rb") as f:
                data = tomli.load(f)
            assert isinstance(data, dict)
        except FileNotFoundError:
            return cls()

        try:
            out = decode(cls, data.get("tool", {}).get("gantry", {}))
        except Exception as e:
            raise ConfigurationError(
                f"Failed to decode gantry config from {path} ({type(e).__name__}: {e})"
            ) from e

        out.path = path
        return out

    def _get_value_string(self, value: Any) -> str:
        if isinstance(value, bool):
            return str(value).lower()
        else:
            return f"'{value}'"

    def get_help_string_for_default(
        self,
        field: Literal[
            "workspace", "gh_token_secret", "budget", "log_level", "quiet", "callback_modules"
        ],
        default: Any = None,
    ) -> str:
        value = getattr(self, field)
        if value is None and default is None:
            return ""
        elif value is None:
            return f"Defaults to {self._get_value_string(default)}."
        elif self.path is not None:
            return f"Defaults to {self._get_value_string(value)} (from config in {self.path})."
        else:
            return f"Defaults to {self._get_value_string(value)} (from config)."


_GLOBAL_CONFIG: GantryConfig | None = None


def get_global_config() -> GantryConfig:
    global _GLOBAL_CONFIG
    if _GLOBAL_CONFIG is None:
        _GLOBAL_CONFIG = GantryConfig.load()
    return _GLOBAL_CONFIG



from dataclasses import dataclass
from typing import Any

import requests
from beaker import BeakerDataset, BeakerJob

from .callback import Callback


@Callback.register("slack")
@dataclass(kw_only=True)
class SlackCallback(Callback):
    webhook_url: str

    @property
    def workload_name(self) -> str:
        return f"{self.beaker.user_name}/{self.workload.experiment.name}"

    @property
    def workload_url(self) -> str:
        return self.beaker.workload.url(self.workload)

    def on_start(self, job: BeakerJob):
        del job
        requests.post(
            self.webhook_url,
            json={
                "text": f":check: Workload <{self.workload_url}|*{self.workload_name}*> has started! :runner:"
            },
        )

    def on_start_timeout(self, job: BeakerJob):
        del job
        requests.post(
            self.webhook_url,
            json={
                "text": f":warning: Workload <{self.workload_url}|*{self.workload_name}*> failed to start in time!"
            },
        )

    def on_timeout(self, job: BeakerJob):
        del job
        requests.post(
            self.webhook_url,
            json={
                "text": f":warning: Workload <{self.workload_url}|*{self.workload_name}*> failed to complete in time!"
            },
        )

    def on_inactive_timeout(self, job: BeakerJob):
        del job
        requests.post(
            self.webhook_url,
            json={
                "text": f":zzz: Workload <{self.workload_url}|*{self.workload_name}*> appears to be inactive!"
            },
        )

    def on_inactive_soft_timeout(self, job: BeakerJob):
        del job
        requests.post(
            self.webhook_url,
            json={
                "text": f":zzz: Workload <{self.workload_url}|*{self.workload_name}*> appears to be inactive!"
            },
        )

    def on_preemption(self, job: BeakerJob):
        del job
        requests.post(
            self.webhook_url,
            json={
                "text": f":warning: Workload <{self.workload_url}|*{self.workload_name}*> was preempted!"
            },
        )

    def on_cancellation(self, job: BeakerJob | None):
        del job
        requests.post(
            self.webhook_url,
            json={
                "text": f":warning: Workload <{self.workload_url}|*{self.workload_name}*> was canceled!"
            },
        )

    def on_failure(
        self,
        job: BeakerJob,
        *,
        metrics: dict[str, Any] | None = None,
        results_ds: BeakerDataset | None = None,
    ):
        del job, metrics, results_ds
        requests.post(
            self.webhook_url,
            json={
                "text": f":check-failed: Workload <{self.workload_url}|*{self.workload_name}*> failed!"
            },
        )

    def on_success(
        self,
        job: BeakerJob,
        *,
        metrics: dict[str, Any] | None = None,
        results_ds: BeakerDataset | None = None,
    ):
        del job, metrics, results_ds
        requests.post(
            self.webhook_url,
            json={
                "text": f":check: Workload <{self.workload_url}|*{self.workload_name}*> succeeded!"
            },
        )



import concurrent.futures
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from pathlib import Path

import click
from beaker import Beaker, BeakerJob, BeakerTask, BeakerWorkload

from .. import beaker_utils, utils
from ..aliases import PathOrStr
from ..exceptions import ConfigurationError
from .main import CLICK_COMMAND_DEFAULTS, main


@main.command(**CLICK_COMMAND_DEFAULTS)
@click.argument("workload", nargs=1, type=str)
@click.option("-r", "--replica", type=int, help="""The replica rank to pull logs from.""")
@click.option("--task", "task_name", type=str, help="""The name of task to pull logs from.""")
@click.option("--run", type=int, help="""The run number to pull logs from.""")
@click.option(
    "-a",
    "--all",
    "all_tasks",
    is_flag=True,
    help="""Pull logs from all tasks/replicas.
    This requires an output directory (e.g. --output=~/Downloads/).""",
)
@click.option("-n", "--head", type=int, help="""Only download the first N lines.""")
@click.option("-t", "--tail", type=int, help="""Tail this many lines.""")
@click.option("-s", "--since", type=str, help="""Only pull logs since this time (ISO format).""")
@click.option(
    "-f", "--follow", is_flag=True, help="""Continue streaming logs for the duration of the job."""
)
@click.option("-o", "--output", type=str, help="""A file or directory to download the logs to.""")
def logs(
    workload: str,
    replica: int | None = None,
    task_name: int | None = None,
    run: int | None = None,
    all_tasks: bool = False,
    head: int | None = None,
    tail: int | None = None,
    since: str | None = None,
    follow: bool = False,
    output: PathOrStr | None = None,
):
    """
    Download the logs from a workload.
    """
    if sum([replica is not None, task_name is not None, all_tasks]) > 1:
        raise ConfigurationError("--replica, --task, and --all are mutually exclusive")

    if all_tasks and follow:
        raise ConfigurationError("--all and --follow are mutually exclusive")

    if run is not None and run < 1:
        raise ConfigurationError("--run must be at least 1")

    if tail and since:
        raise ConfigurationError("--tail and --since are mutually exclusive")

    if head and tail:
        raise ConfigurationError("--head and --tail are mutually exclusive")

    if output is not None:
        output = Path(output)

    since_dt: datetime | None = None
    if since is not None:
        since_dt = datetime.fromisoformat(since.replace("Z", "+00:00")) - timedelta(milliseconds=1)

    with beaker_utils.init_client(ensure_workspace=False) as beaker:
        wl = beaker.workload.get(workload)
        tasks = list(wl.experiment.tasks)

        job: BeakerJob | None = None
        task: BeakerTask | None = None
        out_path: Path | None = None
        if all_tasks:
            if output is None:
                raise ConfigurationError(
                    "An --output directory must be provided when pulling logs from all tasks/replicas"
                )
            elif not output.is_dir() and output.suffix:
                raise ConfigurationError(
                    "--output must be a directory when pulling logs from all tasks/replicas"
                )

            if wl.experiment.id and wl.experiment.id not in str(output):
                output = output / wl.experiment.id

            utils.print_stdout(f"Pulling logs from {len(tasks):,d} tasks/replicas...")
            with ThreadPoolExecutor() as executor:
                futures = []
                for task in tasks:
                    out_path = _resolve_output_file_path(output, task, run)
                    future = executor.submit(
                        _resolve_job_and_download_logs,
                        beaker,
                        wl=wl,
                        task=task,
                        run=run,
                        head=head,
                        tail=tail,
                        since=since_dt,
                        out_path=out_path,
                    )
                    futures.append(future)

                num_log_files_written = 0
                min_lines_logged: int | None = None
                file_with_min_lines: Path | None = None
                max_lines_logged: int | None = None
                file_with_max_lines: Path | None = None
                for future in concurrent.futures.as_completed(futures):
                    task, job, out_path, n_lines = future.result()
                    assert task is not None
                    if job is None:
                        utils.print_stderr(
                            f"[yellow]Job for task '{task.name}' has not started yet[/]"
                        )
                    else:
                        num_log_files_written += 1
                        if min_lines_logged is None or n_lines < min_lines_logged:
                            min_lines_logged = n_lines
                            file_with_min_lines = out_path
                        if max_lines_logged is None or n_lines > max_lines_logged:
                            max_lines_logged = n_lines
                            file_with_max_lines = out_path

            if num_log_files_written > 0:
                utils.print_stdout(f"Logs saved to [cyan]{output}[/]")

            if (
                num_log_files_written > 1
                and min_lines_logged is not None
                and file_with_min_lines is not None
            ):
                utils.print_stdout(
                    f"Log file with fewest lines: [cyan]{file_with_min_lines}[/] ({min_lines_logged:,d} lines)"
                )

            if (
                num_log_files_written > 1
                and max_lines_logged is not None
                and file_with_max_lines is not None
            ):
                utils.print_stdout(
                    f"Log file with most lines: [cyan]{file_with_max_lines}[/] ({max_lines_logged:,d} lines)"
                )
        else:
            if replica is not None:
                for task in tasks:
                    job = beaker_utils.get_job(beaker, wl, task, run=run)
                    if job is not None and job.system_details.replica_group_details.rank == replica:
                        break
                else:
                    raise ConfigurationError(f"Invalid replica rank '{replica}'")
            elif task_name is not None:
                for task in tasks:
                    if task.name == task_name:
                        job = beaker_utils.get_job(beaker, wl, task, run=run)
                        break
                else:
                    raise ConfigurationError(f"Invalid task name '{task_name}'")
            else:
                task = tasks[0]
                job = beaker_utils.get_job(beaker, wl, task, run=run)

            if job is None:
                utils.print_stderr("[yellow]Experiment has not started yet[/]")
                return

            if output is not None:
                out_path = _resolve_output_file_path(output, task, run)

            utils.print_stdout(f"Pulling logs from job '{job.id}' for task '{task.name}'...")
            beaker_utils.download_logs(
                beaker,
                job,
                head_lines=head,
                tail_lines=tail,
                follow=follow,
                since=since_dt,
                out_path=out_path,
            )

            if out_path is not None:
                utils.print_stdout(f"Logs saved to [cyan]{out_path}[/]")


def _resolve_output_file_path(output: Path, task: BeakerTask, run: int | None) -> Path:
    if output.is_dir() or not output.suffix:
        if run is not None:
            return output / f"{task.name}-run-{run}.log"
        else:
            return output / f"{task.name}.log"
    else:
        return output


def _resolve_job_and_download_logs(
    beaker: Beaker,
    *,
    wl: BeakerWorkload,
    task: BeakerTask,
    run: int | None,
    out_path: Path,
    head: int | None = None,
    tail: int | None = None,
    since: datetime | None = None,
) -> tuple[BeakerTask, BeakerJob | None, Path, int]:
    job = beaker_utils.get_job(beaker, wl, task, run=run)
    if job is None:
        return task, None, out_path, 0
    job, n_lines = beaker_utils.download_logs(
        beaker,
        job,
        head_lines=head,
        tail_lines=tail,
        follow=False,
        since=since,
        out_path=out_path,
    )
    return task, job, out_path, n_lines



import json
import logging
import platform
import re
import time
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from functools import cache
from pathlib import Path
from typing import Any, Literal, Optional, cast

import rich
from rich.console import Console, ConsoleRenderable
from rich.text import Text
from rich.traceback import Traceback

from . import constants
from .exceptions import *
from .version import VERSION

log = logging.getLogger()


VERSION_CHECK_INTERVAL = 12 * 3600  # 12 hours
DEFAULT_INTERNAL_CONFIG_LOCATION: Path | None = None
try:
    DEFAULT_INTERNAL_CONFIG_LOCATION = Path.home() / ".beaker" / ".beaker-gantry.json"
except RuntimeError:
    # Can't locate home directory.
    pass

CLI_MODE = False


def enable_cli_mode():
    global CLI_MODE
    CLI_MODE = True


def is_cli_mode() -> bool:
    return CLI_MODE


def is_interactive_terminal() -> bool:
    return rich.get_console().is_interactive and rich.get_console().is_terminal


@dataclass
class InternalConfig:
    version_checked: float | None = None

    @classmethod
    def load(cls) -> Optional["InternalConfig"]:
        path = DEFAULT_INTERNAL_CONFIG_LOCATION
        if path is None:
            return None
        elif path.is_file():
            try:
                with open(path, "r") as f:
                    return cls(**json.load(f))
            except Exception as e:
                log.exception(
                    f"Loading internal config failed with: {type(e).__name__}: {e}\n"
                    f"If this happens consistently then the config file at '{path}' is probably corrupted and should be deleted.\n"
                    "You can also skip this check by adding the CLI flag '--no-check-for-upgrades' or "
                    "by setting the environment variable 'GANTRY_CHECK_FOR_UPGRADES=0'."
                )
                return None
        else:
            return cls()

    def save(self):
        path = DEFAULT_INTERNAL_CONFIG_LOCATION
        if path is None:
            return None
        else:
            path.parent.mkdir(exist_ok=True, parents=True)
            with open(path, "w") as f:
                json.dump(asdict(self), f)


def unique_suffix(max_chars: int = 7) -> str:
    import uuid

    return str(uuid.uuid4())[:max_chars]


def unique_name() -> str:
    import petname

    return cast(str, petname.generate()) + "-" + unique_suffix()


def get_local_python_version() -> str:
    return ".".join(platform.python_version_tuple()[:-1])


def fmt_opt(option: str) -> str:
    if is_cli_mode():
        option = option.strip("-").replace("_", "-")
        return f"'--{option}'"
    else:
        option = option.split("/")[0].strip("- ").replace("-", "_")
        return f"'{option}'"


@cache
def stderr_console() -> Console:
    return Console(stderr=True)


def print_stderr(*args, **kwargs):
    stderr_console().print(*args, **kwargs, highlight=False)


def print_stdout(*args, highlight: bool = False, markup: bool = True, **kwargs):
    rich.get_console().print(*args, **kwargs, highlight=highlight, markup=markup)


def print_exception(*args, **kwargs):
    stderr_console().print_exception(*args, **kwargs)


def parse_timedelta(dur: str) -> timedelta:
    dur_normalized = dur.replace(" ", "").lower()
    if not re.match(r"^(([0-9.e-]+)([a-z]*))+$", dur_normalized):
        raise ValueError(f"invalid duration string '{dur}'")

    seconds = 0.0
    for match in re.finditer(r"([0-9.e-]+)([a-z]*)", dur_normalized):
        value_str, unit = match.group(1), match.group(2)
        try:
            value = float(value_str)
        except ValueError:
            raise ValueError(f"invalid duration string '{dur}'")

        if not unit:
            # assume seconds
            unit = "s"

        if unit in ("ns", "nanosecond", "nanoseconds"):
            # nanoseconds
            seconds += value / 1_000_000_000
        elif unit in ("µs", "microsecond", "microseconds"):
            seconds += value / 1_000_000
        elif unit in ("ms", "millisecond", "milliseconds"):
            # milliseconds
            seconds += value / 1_000
        elif unit in ("s", "sec", "second", "seconds"):
            # seconds
            seconds += value
        elif unit in ("m", "min", "minute", "minutes"):
            # minutes
            seconds += value * 60
        elif unit in ("h", "hr", "hour", "hours"):
            # hours
            seconds += value * 3_600
        elif unit in ("d", "day", "days"):
            # days
            seconds += value * 86_400
        else:
            raise ValueError(f"invalid duration string '{dur}'")

    return timedelta(seconds=seconds)


def format_timedelta(
    td: timedelta | float, resolution: Literal["seconds", "minutes"] = "seconds"
) -> str:
    if isinstance(td, float):
        td = timedelta(seconds=td)

    def format_value_and_unit(value: int, unit: str) -> str:
        if value == 1:
            return f"{value} {unit}"
        else:
            return f"{value} {unit}s"

    seconds = int(td.total_seconds())
    if resolution == "minutes":
        seconds = round(seconds / 60) * 60

    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)

    parts = []
    if days:
        parts.append(format_value_and_unit(days, "day"))
    if hours:
        parts.append(format_value_and_unit(hours, "hour"))
    if minutes:
        parts.append(format_value_and_unit(minutes, "minute"))
    if seconds:
        parts.append(format_value_and_unit(seconds, "second"))

    if not parts:
        if resolution == "minutes":
            parts.append("under 1 minute")
        else:
            parts.append("under 1 second")

    return ", ".join(parts)


def maybe_truncate_text(s: str, n: int) -> str:
    if len(s) <= n:
        return s
    else:
        return s[: n - 1] + "…"


def check_for_upgrades(force: bool = False):
    config = InternalConfig.load()
    if (
        not force
        and config is not None
        and config.version_checked is not None
        and (time.time() - config.version_checked <= VERSION_CHECK_INTERVAL)
    ):
        return

    import packaging.version
    import requests

    try:
        response = requests.get(
            "https://pypi.org/simple/beaker-gantry",
            headers={"Accept": "application/vnd.pypi.simple.v1+json"},
            timeout=2,
        )
        if response.ok:
            latest_version = packaging.version.parse(response.json()["versions"][-1])
            current_version = packaging.version.parse(VERSION)
            if latest_version > current_version and (
                not latest_version.is_prerelease or current_version.is_prerelease
            ):
                print_stderr(
                    f":warning: [yellow]You're using [b]gantry v{VERSION}[/], "
                    f"but a newer version ([b]v{latest_version}[/]) is available: "
                    f"https://github.com/allenai/beaker-gantry/releases/tag/v{latest_version}[/]\n"
                    f"[yellow i]You can upgrade by running:[/] pip install --upgrade beaker-gantry beaker-py\n",
                )
            if config is not None:
                config.version_checked = time.time()
                config.save()
    except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
        pass


def import_module(module_name: str):
    import importlib

    return importlib.import_module(module_name)


def highlight_pattern(s: str, pattern: str) -> str:
    match = s.lower()
    pattern = pattern.lower()
    start_offset = 0
    while (match_start := match.find(pattern, start_offset)) > -1:
        match_str = f"[b green]{pattern.upper()}[/]"
        s = s[:match_start] + match_str + s[match_start + len(pattern) :]
        start_offset = match_start + len(match_str)
        match = s.lower()
    return s


def replace_tags(contents: bytes) -> bytes:
    tag_start = contents.find(b"${{")
    while tag_start != -1:
        tag_end = contents.find(b"}}") + 2
        tag = contents[tag_start:tag_end]
        constant_name = tag.split(b" ")[1].decode()
        contents = contents.replace(tag, getattr(constants, constant_name).encode())  # type: ignore
        tag_start = contents.find(b"${{", tag_end)
    assert b"${{" not in contents
    return contents


class RichHandler(logging.Handler):
    """
    A simplified version of rich.logging.RichHandler from
    https://github.com/Textualize/rich/blob/master/rich/logging.py
    """

    def __init__(
        self,
        *,
        level: int | str = logging.NOTSET,
        console: Optional[Console] = None,
        markup: bool = False,
    ) -> None:
        super().__init__(level=level)
        self.console = console or rich.get_console()
        self.markup = markup

    def emit(self, record: logging.LogRecord) -> None:
        try:
            if hasattr(record.msg, "__rich__") or hasattr(record.msg, "__rich_console__"):
                self.console.print(record.msg)
            else:
                msg: Any = record.msg
                if isinstance(record.msg, str):
                    msg = self.render_message(record=record, message=record.getMessage())
                renderables = [
                    self.get_time_text(record),
                    self.get_level_text(record),
                    self.get_location_text(record),
                    msg,
                ]
                if record.exc_info is not None:
                    tb = Traceback.from_exception(*record.exc_info)  # type: ignore
                    renderables.append(tb)
                self.console.print(*renderables)
        except Exception:
            self.handleError(record)

    def render_message(self, *, record: logging.LogRecord, message: str) -> ConsoleRenderable:
        use_markup = getattr(record, "markup", self.markup)
        message_text = Text.from_markup(message) if use_markup else Text(message)
        return message_text

    def get_time_text(self, record: logging.LogRecord) -> Text:
        log_time = datetime.fromtimestamp(record.created)
        time_str = log_time.strftime("[%Y-%m-%d %X]")
        return Text(time_str, style="log.time", end=" ")

    def get_level_text(self, record: logging.LogRecord) -> Text:
        level_name = record.levelname
        level_text = Text.styled(level_name.ljust(8), f"logging.level.{level_name.lower()}")
        level_text.style = "log.level"
        level_text.end = " "
        return level_text

    def get_location_text(self, record: logging.LogRecord) -> Text:
        name_and_line = f"{record.name}:{record.lineno}" if record.name != "root" else "root"
        text = f"[{name_and_line}]"  # type: ignore
        return Text(text, style="log.path")



import dataclasses
from dataclasses import dataclass
from typing import Any

from beaker import (
    Beaker,
    BeakerDataset,
    BeakerExperimentSpec,
    BeakerJob,
    BeakerWorkload,
)
from dataclass_extensions import Registrable

from ..exceptions import GantryInterruptWorkload
from ..git_utils import GitRepoState


@dataclass(kw_only=True)
class Callback(Registrable):
    """
    Base class for gantry callbacks. Callbacks provide a way to hook into gantry's launch
    loop to customize behavior on certain events.
    """

    _beaker: Beaker | None = dataclasses.field(repr=False, init=False, default=None)
    _git_repo: GitRepoState | None = dataclasses.field(repr=False, init=False, default=None)
    _spec: BeakerExperimentSpec | None = dataclasses.field(repr=False, init=False, default=None)
    _workload: BeakerWorkload | None = dataclasses.field(repr=False, init=False, default=None)

    @property
    def beaker(self) -> Beaker:
        """
        A beaker client that can be accessed after :meth:`attach()` is called.
        """
        if self._beaker is None:
            raise RuntimeError("Callback has not been attached to a gantry workload yet")
        return self._beaker

    @property
    def git_repo(self) -> GitRepoState:
        """
        The git repo state that can be accessed after :meth:`attach()` is called.
        """
        if self._git_repo is None:
            raise RuntimeError("Callback has not been attached to a gantry workload yet")
        return self._git_repo

    @property
    def spec(self) -> BeakerExperimentSpec:
        """
        The experiment spec that can be accessed after :meth:`attach()` is called.
        """
        if self._spec is None:
            raise RuntimeError("Callback has not been attached to a gantry workload yet")
        return self._spec

    @property
    def workload(self) -> BeakerWorkload:
        """
        The workload that can be accessed after :meth:`attach()` is called.
        """
        if self._workload is None:
            raise RuntimeError("Callback has not been attached to a gantry workload yet")
        return self._workload

    def interrupt_workload(self):
        """Cancels the active workload."""
        raise GantryInterruptWorkload(f"workload interrupted by callback {self.__class__.__name__}")

    def attach(
        self,
        *,
        beaker: Beaker,
        git_repo: GitRepoState,
        spec: BeakerExperimentSpec,
        workload: BeakerWorkload,
    ):
        """
        Runs when a callback is attached to the workload.
        """
        self._beaker = beaker
        self._git_repo = git_repo
        self._spec = spec
        self._workload = workload

    def detach(self):
        """
        Runs when a callback is detached from the workload.
        """
        self._beaker = None
        self._git_repo = None
        self._spec = None
        self._workload = None

    def on_start(self, job: BeakerJob):
        """
        Runs when a job for the workload starts.
        """
        del job

    def on_log(self, job: BeakerJob, log_line: str, log_time: float):
        """
        Runs when a new log event is received from the workload.
        """
        del job, log_line, log_time

    def on_no_new_logs(self, job: BeakerJob):
        """
        Periodically runs when no new logs have been received from the workload recently.
        Runs at most once per second.
        """
        del job

    def on_start_timeout(self, job: BeakerJob):
        """
        Runs when the active job for the workload hits the configured start timeout before starting.
        """
        del job

    def on_timeout(self, job: BeakerJob):
        """
        Runs when the active job for the workload hits the configured timeout before completing.
        """
        del job

    def on_inactive_timeout(self, job: BeakerJob):
        """
        Runs when the active job for the workload hits the configured inactive timeout.
        """
        del job

    def on_inactive_soft_timeout(self, job: BeakerJob):
        """
        Runs when the active job for the workload hits the configured inactive hard timeout.
        """
        del job

    def on_preemption(self, job: BeakerJob):
        """
        Runs when the active job for the workload is preempted.
        """
        del job

    def on_cancellation(self, job: BeakerJob | None):
        """
        Runs when the active job for the workload is canceled by the user, either directly or because.
        a timeout was reached.
        """
        del job

    def on_failure(
        self,
        job: BeakerJob,
        *,
        metrics: dict[str, Any] | None = None,
        results_ds: BeakerDataset | None = None,
    ):
        """
        Runs when the active job for the workload fails.
        """
        del job, metrics, results_ds

    def on_success(
        self,
        job: BeakerJob,
        *,
        metrics: dict[str, Any] | None = None,
        results_ds: BeakerDataset | None = None,
    ):
        """
        Runs when the active job for the workload succeeds.
        """
        del job, metrics, results_ds



import concurrent.futures
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta, timezone
from itertools import islice
from typing import Generator, Iterable

import click
import rich
from beaker import (
    Beaker,
    BeakerSortOrder,
    BeakerTask,
    BeakerWorkload,
    BeakerWorkloadStatus,
    BeakerWorkloadType,
)
from beaker.exceptions import BeakerGroupNotFound
from rich.table import Table

from .. import beaker_utils, utils
from ..config import get_global_config
from ..exceptions import ConfigurationError
from .main import CLICK_COMMAND_DEFAULTS, main


class Defaults:
    limit = 10
    max_age = 7


@main.command(name="list", **CLICK_COMMAND_DEFAULTS)
@click.option(
    "-w",
    "--workspace",
    type=str,
    help=f"""The Beaker workspace to pull experiments from. {get_global_config().get_help_string_for_default('workspace')}""",
)
@click.option("-g", "--group", type=str, help="""The Beaker group to pull experiments from.""")
@click.option(
    "-l",
    "--limit",
    type=int,
    default=Defaults.limit,
    help=f"Limit the number of experiments to display. Default: {Defaults.limit}",
)
@click.option(
    "-a",
    "--author",
    type=str,
    help="Filter by author. Tip: use '--me' instead to show your own experiments.",
)
@click.option(
    "--me",
    is_flag=True,
    help="Only show your own experiments. Mutually exclusive with '--author'.",
)
@click.option(
    "-s",
    "--status",
    type=click.Choice([x.name for x in BeakerWorkloadStatus]),
    help="Filter by status. Multiple allowed.",
    multiple=True,
)
@click.option(
    "-t",
    "--text",
    "name_or_description",
    type=str,
    help="Filter by text in the experiment name or description.",
)
@click.option(
    "--max-age",
    "max_age_str",
    type=str,
    default=Defaults.max_age,
    help=f"""Maximum age of experiments (in days, by default).
    Can also be specified as a duration such as '5m', '2h', etc.
    Default: {Defaults.max_age} days.""",
)
@click.option(
    "--all",
    "show_all",
    is_flag=True,
    help="Show all experiments, not just onces submitted through Gantry.",
)
def list_cmd(
    workspace: str | None = None,
    group: str | None = None,
    limit: int = Defaults.limit,
    author: str | None = None,
    me: bool = False,
    status: list[str] | None = None,
    name_or_description: str | None = None,
    max_age_str: str = str(Defaults.max_age),
    show_all: bool = False,
):
    """
    List workloads within a workspace or group.
    This will only show workloads launched with Gantry by default, unless '--all' is specified.
    """
    max_age: timedelta
    try:
        max_age = timedelta(days=int(max_age_str))
    except ValueError:
        max_age = utils.parse_timedelta(max_age_str)

    with beaker_utils.init_client(ensure_workspace=False) as beaker:
        table = Table(title="Experiments", show_lines=True)
        table.add_column("Workload", justify="left", no_wrap=True)
        table.add_column("Author", justify="left", style="blue", no_wrap=True)
        table.add_column("Created", justify="left", no_wrap=True)
        table.add_column("Tasks")

        if me:
            if author is not None:
                raise ConfigurationError("--me and -a/--author are mutually exclusive.")
            else:
                author = beaker.user_name

        status_msg = "[i]collecting experiments...[/]"
        with rich.get_console().status(status_msg, spinner="point", speed=0.8) as status_:
            with ThreadPoolExecutor() as executor:
                for wl, tasks in islice(
                    iter_workloads(
                        beaker,
                        executor=executor,
                        workspace=workspace,
                        group=group,
                        author=author,
                        statuses=status,
                        name_or_description=name_or_description,
                        max_age=max_age,
                        show_all=show_all,
                        limit=None if not show_all else limit,
                    ),
                    limit,
                ):
                    status_.update(f"{status_msg} [cyan]{wl.experiment.name}[/]")

                    task_status_futures = []
                    for task in tasks:
                        task_status_futures.append(executor.submit(format_task, beaker, wl, task))

                    task_statuses = {}
                    for task_future in concurrent.futures.as_completed(task_status_futures):
                        task, task_status = task_future.result()
                        task_statuses[task.id] = task_status
                        status_.update(f"{status_msg} [cyan]{wl.experiment.name} ❯ [/]{task.name}")

                    table.add_row(
                        f"[b cyan]{wl.experiment.name}[/]\n"
                        f"[blue u]{beaker.workload.url(wl)}[/]\n"
                        f"{utils.maybe_truncate_text(wl.experiment.description, max(len(beaker.workload.url(wl)), len(wl.experiment.name)))}".strip(),
                        beaker.user.get(wl.experiment.author_id).name,
                        wl.experiment.created.ToDatetime(timezone.utc)
                        .astimezone(tz=None)
                        .strftime("%I:%M %p on %a, %b %-d"),
                        "\n".join(task_statuses[task.id] for task in tasks),
                    )

        utils.print_stdout(table)


def is_gantry_workload(beaker: Beaker, wl: BeakerWorkload) -> bool:
    spec = beaker.experiment.get_spec(wl.experiment)
    for task_spec in spec.tasks:
        for env_var in task_spec.env_vars or []:
            if env_var.name == "GANTRY_VERSION":
                return True
    return False


def iter_workloads(
    beaker: Beaker,
    *,
    executor: ThreadPoolExecutor,
    workspace: str | None,
    group: str | None,
    author: str | None,
    statuses: list[str] | None,
    name_or_description: str | None,
    max_age: timedelta,
    show_all: bool,
    limit: int | None = None,
) -> Generator[tuple[BeakerWorkload, Iterable[BeakerTask]], None, None]:
    created_after = datetime.now(tz=timezone.utc).astimezone() - max_age
    author_id = None if author is None else beaker.user.get(author)
    workload_statuses = (
        None if not statuses else [BeakerWorkloadStatus.from_any(s) for s in statuses]
    )

    if group is not None:
        beaker_group = beaker_utils.resolve_group(
            beaker, group, workspace, fall_back_to_default_workspace=False
        )
        if beaker_group is None:
            raise BeakerGroupNotFound(group)

        # Will have to sort workloads manually by creation time, so we gather them all first.
        workload_futures = []
        for task_metrics in beaker.group.list_task_metrics(beaker_group):
            workload_futures.append(
                executor.submit(beaker.workload.get, task_metrics.experiment_id)
            )

        workloads = []
        for future in concurrent.futures.as_completed(workload_futures):
            wl = future.result()

            # Filter out non-gantry experiments.
            if not show_all and not is_gantry_workload(beaker, wl):
                continue

            if author_id is not None and wl.experiment.author_id != author_id:
                continue

            if workload_statuses is not None and wl.status not in workload_statuses:
                continue

            if wl.experiment.created.ToDatetime(timezone.utc) < created_after:
                continue

            if (
                name_or_description is not None
                and name_or_description not in wl.experiment.name
                and name_or_description not in wl.experiment.description
            ):
                continue

            workloads.append(wl)

        workloads.sort(key=lambda wl: wl.experiment.created.ToMilliseconds(), reverse=True)
        for wl in workloads:
            yield wl, wl.experiment.tasks
    else:
        if workspace is None:
            workspace = get_global_config().workspace

        for wl in beaker.workload.list(
            workspace=None if workspace is None else beaker.workspace.get(workspace),
            author=author_id,
            created_after=created_after,
            workload_type=BeakerWorkloadType.experiment,
            statuses=workload_statuses,
            name_or_description=name_or_description,
            sort_order=BeakerSortOrder.descending,
            sort_field="created",
            limit=limit,
        ):
            # Filter out non-gantry experiments.
            if not show_all and not is_gantry_workload(beaker, wl):
                continue

            yield wl, wl.experiment.tasks


def get_status(beaker: Beaker, wl: BeakerWorkload, task: BeakerTask) -> BeakerWorkloadStatus | None:
    job = beaker.workload.get_latest_job(wl, task=task)
    if job is not None:
        return BeakerWorkloadStatus.from_any(job.status.status)
    else:
        return None


def format_task(beaker: Beaker, wl: BeakerWorkload, task: BeakerTask) -> tuple[BeakerTask, str]:
    style = "i"
    status = get_status(beaker, wl, task)
    if status in (
        BeakerWorkloadStatus.running,
        BeakerWorkloadStatus.succeeded,
        BeakerWorkloadStatus.uploading_results,
        BeakerWorkloadStatus.ready_to_start,
        BeakerWorkloadStatus.initializing,
    ):
        style += " green"
    elif status in (
        BeakerWorkloadStatus.submitted,
        BeakerWorkloadStatus.queued,
        BeakerWorkloadStatus.canceled,
        BeakerWorkloadStatus.stopping,
    ):
        style += " yellow"
    elif status == BeakerWorkloadStatus.failed:
        style += " red"

    status_str = "unknown" if status is None else status.name

    return task, f"[i]{task.name}[/] ([{style}]{status_str}[/])"



import logging
import os
import sys
import time
from contextlib import ExitStack
from pathlib import Path
from queue import Empty, Queue
from threading import Event, Thread
from typing import Literal, Sequence, Type

import rich
from beaker import (
    Beaker,
    BeakerCluster,
    BeakerGroup,
    BeakerJob,
    BeakerSortOrder,
    BeakerTask,
    BeakerTaskResources,
    BeakerWorkload,
)
from beaker.exceptions import (
    BeakerExperimentConflict,
    BeakerGroupNotFound,
    BeakerImageNotFound,
    BeakerSecretNotFound,
)
from rich import prompt
from rich.status import Status

from . import beaker_utils, constants, utils
from .aliases import PathOrStr
from .callbacks import Callback
from .exceptions import *
from .git_utils import GitRepoState

log = logging.getLogger(__name__)


def launch_experiment(
    args: Sequence[str],
    name: str | None = None,
    description: str | None = None,
    task_name: str = "main",
    workspace: str | None = None,
    group_names: Sequence[str] | None = None,
    clusters: Sequence[str] | None = None,
    gpu_types: Sequence[str] | None = None,
    interconnect: Literal["ib", "tcpxo"] | None = None,
    tags: Sequence[str] | None = None,
    hostnames: Sequence[str] | None = None,
    beaker_image: str | None = None,
    docker_image: str | None = None,
    cpus: float | None = None,
    gpus: int | None = None,
    memory: str | None = None,
    shared_memory: str | None = None,
    datasets: Sequence[str] | None = None,
    gh_token_secret: str = constants.GITHUB_TOKEN_SECRET,
    conda_file: PathOrStr | None = None,
    conda_env: str | None = None,
    python_manager: Literal["uv", "conda"] | None = None,
    system_python: bool = False,
    uv_venv: str | None = None,
    uv_extras: Sequence[str] | None = None,
    uv_all_extras: bool | None = None,
    uv_torch_backend: str | None = None,
    uv_cache_dir: str | None = None,
    env_vars: Sequence[str | tuple[str, str]] | None = None,
    env_secrets: Sequence[str | tuple[str, str]] | None = None,
    dataset_secrets: Sequence[str | tuple[str, str]] | None = None,
    mounts: Sequence[str | tuple[str, str]] | None = None,
    weka: Sequence[str | tuple[str, str]] | None = None,
    uploads: Sequence[str | tuple[str, str]] | None = None,
    timeout: int | None = None,
    task_timeout: str | None = None,
    start_timeout: int | None = None,
    inactive_timeout: int | None = None,
    inactive_soft_timeout: int | None = None,
    show_logs: bool | None = None,
    allow_dirty: bool = False,
    dry_run: bool = False,
    yes: bool | None = None,
    save_spec: PathOrStr | None = None,
    priority: str | None = None,
    install: str | None = None,
    no_python: bool = False,
    replicas: int | None = None,
    leader_selection: bool | None = None,
    host_networking: bool | None = None,
    propagate_failure: bool | None = None,
    propagate_preemption: bool | None = None,
    synchronized_start_timeout: str | None = None,
    budget: str | None = None,
    preemptible: bool | None = None,
    retries: int | None = None,
    results: str = constants.RESULTS_DIR,
    runtime_dir: str = constants.RUNTIME_DIR,
    exec_method: Literal["exec", "bash"] = "exec",
    torchrun: bool = False,
    skip_tcpxo_setup: bool = False,
    skip_nccl_setup: bool = False,
    default_python_version: str = utils.get_local_python_version(),
    pre_setup: str | None = None,
    post_setup: str | None = None,
    aws_config_secret: str | None = None,
    aws_credentials_secret: str | None = None,
    google_credentials_secret: str | None = None,
    callbacks: Sequence[Callback] | None = None,
    ref: str | None = None,
    branch: str | None = None,
    remote: str | None = None,
    git_repo: GitRepoState | None = None,
    auto_cancel: bool = False,
    client: Beaker | None = None,
) -> BeakerWorkload | None:
    """
    Launch an experiment on Beaker. Same as the ``gantry run`` command.

    :param cli_mode: Set to ``True`` if this function is being called from a CLI command.
        This mostly affects how certain prompts and messages are displayed.
    """
    if not args:
        if utils.is_cli_mode():
            raise ConfigurationError(
                "[ARGS]... are required! For example:\n  $ gantry run -- python -c 'print(\"Hello, World!\")'"
            )
        else:
            raise ConfigurationError("'args' are required!")

    if torchrun and not gpus:
        raise ConfigurationError(
            f"{utils.fmt_opt('--torchrun')} mode requires {utils.fmt_opt('--gpus')} to be set to a positive integer."
        )

    if replicas is not None:
        if replicas <= 0:
            raise ConfigurationError(
                f"{utils.fmt_opt('--replicas')} must be a positive integer (got {replicas})"
            )
        elif replicas == 1:
            replicas = None

    if yes is None:
        if os.environ.get("GANTRY_GITHUB_TESTING"):
            yes = True
        else:
            yes = False
    if timeout is None:
        timeout = -1 if show_logs else 0
    if show_logs is None:
        show_logs = timeout != 0
    if timeout == 0 and show_logs:
        raise ConfigurationError(
            f"Cannot use {utils.fmt_opt('--show-logs')} with {utils.fmt_opt('--timeout=0')}!"
        )

    task_resources = BeakerTaskResources(
        cpu_count=cpus, gpu_count=gpus, memory=memory, shared_memory=shared_memory
    )

    # Initialize and validate repo state.
    if git_repo is None:
        if remote is not None:
            git_repo = GitRepoState.from_remote(remote, ref=ref, branch=branch)
        else:
            git_repo = GitRepoState.from_env(ref=ref, branch=branch)
    if ref is None and not allow_dirty and git_repo.is_dirty:
        raise DirtyRepoError(
            f"You have uncommitted changes! Use {utils.fmt_opt('--allow-dirty')} to force."
        )

    # Initialize Beaker client and validate workspace.
    with ExitStack() as stack:
        if client is None:
            beaker: Beaker = stack.enter_context(
                beaker_utils.init_client(workspace=workspace, yes=yes)
            )
        else:
            beaker = client

        if beaker_image is None and docker_image is None:
            try:
                beaker.image.get(constants.VERSIONED_DEFAULT_IMAGE)
                beaker_image = constants.VERSIONED_DEFAULT_IMAGE
            except BeakerImageNotFound:
                beaker_image = constants.DEFAULT_IMAGE
        elif beaker_image is not None and docker_image is not None:
            raise ConfigurationError(
                f"{utils.fmt_opt('--beaker-image')} and {utils.fmt_opt('--docker-image')} are mutually exclusive."
            )
        elif beaker_image is not None and beaker_image != constants.DEFAULT_IMAGE:
            try:
                beaker_image = beaker.image.get(beaker_image).id
            except BeakerImageNotFound:
                raise ConfigurationError(f"Beaker image '{beaker_image}' not found")

        if budget is None and not beaker.workspace.get().budget_id:
            budget = prompt.Prompt.ask(
                f"[yellow]Missing {utils.fmt_opt('--budget')} option, "
                "see https://beaker-docs.apps.allenai.org/concept/budgets.html for more information.[/]\n"
                "[i]Please enter the budget account to associate with this experiment[/]",
            )

            if not budget:
                raise ConfigurationError("Budget account must be specified!")

        # Maybe resolve or create group.
        groups: list[BeakerGroup] = []
        if group_names:
            for group_name in group_names:
                group = beaker_utils.resolve_group(beaker, group_name)
                if group is None:
                    if "/" in group_name and not group_name.startswith(f"{beaker.user_name}/"):
                        raise BeakerGroupNotFound(group_name)

                    if prompt.Confirm.ask(
                        f"Group [green]{group_name}[/] not found in workspace, would you like to create this group?"
                    ):
                        group_name = group_name.split("/", 1)[-1]
                        group = beaker.group.create(group_name)
                        utils.print_stdout(
                            f"Group created: [blue u]{beaker_utils.group_url(beaker, group)}[/]"
                        )
                    else:
                        utils.print_stderr("[yellow]canceled[/]")
                        sys.exit(1)

                assert group is not None
                groups.append(group)

        # Get the entrypoint dataset.
        entrypoint_dataset = beaker_utils.ensure_entrypoint_dataset(beaker, budget)

        # Validate the input datasets.
        datasets_to_use = beaker_utils.ensure_datasets(beaker, *datasets) if datasets else []

        env_var_names: set[str] = set()
        env_vars_to_use = []
        for e in env_vars or []:
            if isinstance(e, tuple):
                env_name, val = e
            else:
                try:
                    env_name, val = e.split("=", 1)
                except ValueError:
                    if e in os.environ:
                        env_name, val = e, os.environ[e]
                    else:
                        raise ConfigurationError(f"Invalid env var: '{e}'")
            if env_name in env_var_names:
                raise ConfigurationError(f"Duplicate env var name: '{env_name}'")
            env_var_names.add(env_name)
            env_vars_to_use.append((env_name, val))
        if uv_cache_dir is not None:
            if "UV_CACHE_DIR" in env_var_names:
                raise ConfigurationError(
                    f"Can't set both {utils.fmt_opt('--uv-cache-dir')} and the env var name 'UV_CACHE_DIR' at the same time."
                )
            env_var_names.add("UV_CACHE_DIR")
            env_vars_to_use.append(("UV_CACHE_DIR", uv_cache_dir))

        secret_names: set[str] = set()
        env_secrets_to_use = []
        for e in env_secrets or []:
            if isinstance(e, tuple):
                env_secret_name, secret = e
            else:
                try:
                    env_secret_name, secret = e.split("=", 1)
                except ValueError:
                    if beaker_utils.secret_exists(beaker, e):
                        env_secret_name = e
                        secret = e
                    elif e in os.environ:
                        env_secret_name = e
                        env_secret_value = os.environ[e]
                        utils.print_stderr(
                            f"[yellow]Taking secret value for '{e}' from environment[/]"
                        )
                        secret = beaker_utils.ensure_secret(
                            beaker, env_secret_name, env_secret_value
                        )
                    else:
                        raise ConfigurationError(f"Invalid env secret: '{e}'")

            if env_secret_name in secret_names:
                raise ConfigurationError(f"Duplicate env secret name: '{env_secret_name}'")
            if env_secret_name in env_var_names:
                raise ConfigurationError(
                    f"Env secret name '{env_secret_name}' conflicts with an env var of the same name"
                )
            secret_names.add(env_secret_name)
            env_secrets_to_use.append((env_secret_name, secret))

        dataset_secrets_to_use = []
        for ds in dataset_secrets or []:
            if isinstance(ds, tuple):
                secret, mount_path = ds
            else:
                try:
                    secret, mount_path = ds.split(":", 1)
                except ValueError:
                    raise ValueError(f"Invalid dataset secret: '{ds}'")
            dataset_secrets_to_use.append((secret, mount_path))

        mounts_to_use = []
        for m in mounts or []:
            if isinstance(m, tuple):
                source, target = m
            else:
                try:
                    source, target = m.split(":", 1)
                except ValueError:
                    raise ValueError(f"Invalid dataset mount: '{m}'")
            mounts_to_use.append((source, target))

        weka_buckets = []
        for m in weka or []:
            if isinstance(m, tuple):
                source, target = m
            else:
                try:
                    source, target = m.split(":", 1)
                except ValueError:
                    raise ValueError(f"Invalid weka mount: '{m}'")
            weka_buckets.append((source, target))

        if weka_buckets and not hostnames and (not tags or "storage:weka" not in tags):
            tags = list(tags or [])
            tags.append("storage:weka")

        uploads_to_use = []
        for f in uploads or []:
            if isinstance(f, tuple):
                source, target = f
            else:
                try:
                    source, target = f.split(":", 1)
                except ValueError:
                    raise ValueError(f"Invalid upload spec: '{f}'")
            uploads_to_use.append((source, target))

        if interconnect is not None:
            tags = list(tags or [])
            tags.append(f"interconnect:{interconnect}")

        # Validate clusters.
        if clusters or gpu_types or tags:
            if hostnames:
                raise ConfigurationError(
                    "Cannot specify hostnames with additional cluster selection constraints"
                )

            constraints = []
            if clusters:
                constraints.append(f'''name matches any of: "{'", "'.join(clusters)}"''')
            if tags:
                constraints.append(f'''has the tag(s): "{'", "'.join(tags)}"''')
            if gpu_types:
                constraints.append(
                    f'''has any one of these GPU types: "{'", "'.join(gpu_types)}"'''
                )

            # Collect all clusters that support batch jobs.
            all_clusters: list[BeakerCluster] = []
            for cl in beaker.cluster.list():
                # If 'max_task_timeout' is set to 0 then tasks are not allowed.
                if cl.HasField("max_task_timeout") and cl.max_task_timeout.ToMilliseconds() == 0:
                    continue
                else:
                    all_clusters.append(cl)

            if not all_clusters:
                raise RuntimeError("Failed to find any clusters that support batch jobs")

            # Maybe filter clusters based on provided patterns.
            if clusters:
                all_clusters = beaker_utils.filter_clusters_by_name(beaker, all_clusters, clusters)

            # Maybe filter based on tags.
            if tags:
                all_clusters = beaker_utils.filter_clusters_by_tags(beaker, all_clusters, tags)

            # Filter based on GPU types.
            if gpu_types:
                all_clusters = beaker_utils.filter_clusters_by_gpu_type(
                    beaker, all_clusters, gpu_types
                )

            if not all_clusters:
                constraints_str = "\n - ".join(constraints)
                raise ConfigurationError(
                    f"Failed to find clusters satisfying the given constraints:\n - {constraints_str}"
                )
            else:
                clusters = [f"{cl.organization_name}/{cl.name}" for cl in all_clusters]

        # Default to preemptible when no cluster has been specified.
        if not clusters and preemptible is None:
            preemptible = True

        # Get / set the GitHub token secret.
        gh_token_secret_to_use: str | None = None
        if not git_repo.is_public and "GITHUB_TOKEN" not in secret_names:
            try:
                beaker.secret.get(gh_token_secret)
            except BeakerSecretNotFound:
                utils.print_stderr(
                    f"[yellow]GitHub token secret '{gh_token_secret}' not found in workspace.[/]\n"
                    f"You can create a suitable GitHub token by going to https://github.com/settings/tokens/new "
                    f"and generating a token with the '\N{ballot box with check} repo' scope."
                )

                if "GITHUB_TOKEN" in os.environ:
                    gh_token = prompt.Prompt.ask(
                        "[i]Please paste your GitHub token here or press [/]ENTER[i] to use your local [/]GITHUB_TOKEN",
                        password=True,
                        default=os.environ["GITHUB_TOKEN"],
                        show_default=False,
                    )
                else:
                    gh_token = prompt.Prompt.ask(
                        "[i]Please paste your GitHub token here[/]",
                        password=True,
                    )

                if not gh_token:
                    raise ConfigurationError("token cannot be empty!")

                beaker.secret.write(gh_token_secret, gh_token)
                utils.print_stdout(
                    f"GitHub token secret uploaded to workspace as '{gh_token_secret}'.\n"
                    f"If you need to update this secret in the future, use the command:\n"
                    f"[i]$ gantry config set-gh-token[/]"
                )

            gh_token_secret_to_use = gh_token_secret

        if aws_config_secret is not None:
            try:
                beaker.secret.get(aws_config_secret)
            except BeakerSecretNotFound:
                raise ConfigurationError(
                    f"AWS config secret '{aws_config_secret}' not found in workspace"
                )
            env_secrets_to_use.append(("GANTRY_AWS_CONFIG", aws_config_secret))

        if aws_credentials_secret is not None:
            try:
                beaker.secret.get(aws_credentials_secret)
            except BeakerSecretNotFound:
                raise ConfigurationError(
                    f"AWS credentials secret '{aws_credentials_secret}' not found in workspace"
                )
            env_secrets_to_use.append(("GANTRY_AWS_CREDENTIALS", aws_credentials_secret))

        if google_credentials_secret is not None:
            try:
                beaker.secret.get(google_credentials_secret)
            except BeakerSecretNotFound:
                raise ConfigurationError(
                    f"Google Cloud credentials secret '{google_credentials_secret}' not found in workspace"
                )
            env_secrets_to_use.append(("GANTRY_GOOGLE_CREDENTIALS", google_credentials_secret))

        if leader_selection is None:
            if replicas and torchrun and host_networking is not False:
                leader_selection = True
            else:
                leader_selection = False

        if host_networking is None:
            if replicas and leader_selection:
                host_networking = True
            else:
                host_networking = False

        if torchrun and replicas and leader_selection:
            if propagate_failure is None:
                propagate_failure = True
            if propagate_preemption is None and propagate_failure:
                propagate_preemption = True
            if synchronized_start_timeout is None:
                synchronized_start_timeout = "5m"

        # Initialize experiment and task spec.
        spec = beaker_utils.build_experiment_spec(
            beaker,
            task_name=task_name,
            clusters=list(clusters or []),
            task_resources=task_resources,
            arguments=list(args),
            entrypoint_dataset=entrypoint_dataset.id,
            git_repo=git_repo,
            budget=budget,
            group_names=[group.full_name for group in groups],
            description=description,
            beaker_image=beaker_image,
            docker_image=docker_image,
            gh_token_secret=gh_token_secret_to_use,
            conda_file=conda_file,
            conda_env=conda_env,
            python_manager=python_manager,
            system_python=system_python,
            uv_venv=uv_venv,
            uv_extras=uv_extras,
            uv_all_extras=uv_all_extras,
            uv_torch_backend=uv_torch_backend,
            datasets=datasets_to_use,
            env=env_vars_to_use,
            env_secrets=env_secrets_to_use,
            dataset_secrets=dataset_secrets_to_use,
            priority=priority,
            install=install,
            no_python=no_python,
            replicas=replicas,
            leader_selection=leader_selection,
            host_networking=host_networking,
            propagate_failure=propagate_failure,
            propagate_preemption=propagate_preemption,
            synchronized_start_timeout=synchronized_start_timeout,
            task_timeout=task_timeout,
            mounts=mounts_to_use,
            weka_buckets=weka_buckets,
            uploads=uploads_to_use,
            hostnames=None if hostnames is None else list(hostnames),
            preemptible=preemptible,
            retries=retries,
            results=results,
            runtime_dir=runtime_dir,
            exec_method=exec_method,
            torchrun=torchrun,
            skip_nccl_setup=skip_nccl_setup or skip_tcpxo_setup,
            default_python_version=default_python_version,
            pre_setup=pre_setup,
            post_setup=post_setup,
        )

        if save_spec:
            if (
                Path(save_spec).is_file()
                and not yes
                and not prompt.Confirm.ask(
                    f"[yellow]The file '{save_spec}' already exists. "
                    f"[i]Are you sure you want to overwrite it?[/][/]"
                )
            ):
                raise KeyboardInterrupt
            spec.to_file(save_spec)
            utils.print_stdout(f"Experiment spec saved to {save_spec}")

        if not name:
            default_name = utils.unique_name()
            if yes:
                name = default_name
            else:
                name = prompt.Prompt.ask(
                    "[i]What would you like to call this experiment?[/]",
                    default=utils.unique_name(),
                )

        if not name:
            raise ConfigurationError("Experiment name cannot be empty!")

        if dry_run:
            rich.get_console().rule("[b]Dry run[/]")

        if groups:
            groups_str = "\n ❯ ".join(
                [
                    f"[cyan]{group.full_name}[/] → [blue u]{beaker_utils.group_url(beaker, group)}[/]"
                    for group in groups
                ]
            )
        else:
            groups_str = ""

        beaker.workspace.get().name
        info_header = (
            f"[b]Workspace:[/] [cyan]{beaker.workspace.get().name}[/] → [blue u]{beaker.workspace.url()}[/]\n"
            + (("[b]Groups:[/]\n ❯ " + groups_str + "\n") if groups else "")
            + f"[b]Commit:[/] [cyan]{git_repo.short_ref}[/] {git_repo.short_commit_message() or ''} → [blue u]{git_repo.ref_url}[/]\n"
            + f"[b]Branch:[/] [cyan]{git_repo.branch}[/]"
            + (f" → [blue u]{git_repo.branch_url}[/]" if git_repo.branch_url else "")
        )

        if dry_run:
            utils.print_stdout(info_header)
            utils.print_stdout(
                f"[b]Name:[/] [cyan]{name}[/]\n[b]Experiment spec:[/]",
                spec.to_json(),
                highlight=True,
            )
            return None

        # Beaker experiment names have to be unique, so we add a random suffix if needed.
        name_prefix = name
        attempts = 0
        while True:
            try:
                workload = beaker.experiment.create(name=name, spec=spec)
                break
            except BeakerExperimentConflict:
                attempts += 1
                if attempts == 5:
                    utils.print_stderr(
                        f"[yellow]Many experiments with the name '{name_prefix}' already exist. Consider using a different name.[/]"
                    )
                name_suffix = utils.unique_suffix(max_chars=4 if attempts < 5 else 7)
                name = f"{name_prefix}-{name_suffix}"

        try:
            info_header = (
                f"[b]Experiment:[/] [cyan]{beaker.user_name}/{workload.experiment.name}[/] → [blue u]{beaker.workload.url(workload)}[/]\n"
                + info_header
            )
            utils.print_stdout(info_header)

            # Initialize and attach callbacks.
            callbacks = list(callbacks) if callbacks is not None else []
            for callback in callbacks:
                callback.attach(
                    beaker=beaker,
                    git_repo=git_repo,
                    spec=spec,
                    workload=workload,
                )

            # Can return right away if timeout is 0.
            if timeout == 0:
                return workload
        except (TermInterrupt, KeyboardInterrupt) as exc:
            utils.print_stderr(f"[red][bold]{exc.__class__.__name__}:[/] [i]{exc}[/][/]")
            if auto_cancel:
                beaker.workload.cancel(workload)
                utils.print_stderr("[yellow]Experiment cancelled.[/]")

        job = follow_workload(
            beaker,
            workload,
            timeout=timeout if timeout > 0 else None,
            start_timeout=start_timeout,
            inactive_timeout=inactive_timeout,
            inactive_soft_timeout=inactive_soft_timeout,
            show_logs=show_logs,
            auto_cancel=auto_cancel,
            callbacks=callbacks,
        )

        try:
            beaker_utils.display_results(
                beaker,
                workload,
                job,
                info_header=info_header if show_logs else None,
                callbacks=callbacks,
            )
        finally:
            for callback in callbacks or []:
                callback.detach()

        return workload


def follow_workload(
    beaker: Beaker,
    workload: BeakerWorkload,
    *,
    job: BeakerJob | None = None,
    task: BeakerTask | None = None,
    timeout: int | None = None,
    start_timeout: int | None = None,
    inactive_timeout: int | None = None,
    inactive_soft_timeout: int | None = None,
    tail: bool = False,
    show_logs: bool = True,
    auto_cancel: bool = False,
    callbacks: Sequence[Callback] | None = None,
) -> BeakerJob:
    """
    Follow a workload until completion while streaming logs to stdout.

    :param task: A specific task in the workload to follow. Defaults to the first task.
    :param timeout: The number of seconds to wait for the workload to complete. Raises a timeout
        error if it doesn't complete in time.
    :param start_timeout: The number of seconds to wait for the workload to start running.
        Raises a timeout error if it doesn't start in time.
    :param inactive_timeout: The number of seconds to wait for new logs before timing out.
        Raises a timeout error if no new logs are produced in time.
    :param inactive_soft_timeout: The number of seconds to wait for new logs before timing out.
        Issues a warning notification if no new logs are produced in time.
    :param tail: Start tailing the logs if a job is already running. Otherwise shows all logs.
    :param show_logs: Set to ``False`` to avoid streaming the logs.
    :param auto_cancel: Set to ``True`` to automatically cancel the workload on timeout or
        or SIGTERM.

    :returns: The finalized :class:`~beaker.types.BeakerJob` from the task being followed.

    :raises ~gantry.exceptions.BeakerJobTimeoutError: If ``timeout`` is set to a positive number
        and the workload doesn't complete in time.
    """
    console = rich.get_console()
    start_time = time.monotonic()
    preempted_job_ids = set()
    on_start_called = set()
    stopped: Event | None = None
    sentinel = object()
    exceptions_to_cancel_on: tuple[Type[BaseException], ...] = (GantryInterruptWorkload,)
    if auto_cancel:
        exceptions_to_cancel_on = exceptions_to_cancel_on + (TermInterrupt, BeakerJobTimeoutError)

    def get_latest_job() -> BeakerJob | None:
        job = beaker.workload.get_latest_job(workload, task=task)
        if job is not None and job.id not in preempted_job_ids:
            return job
        else:
            return None

    def fill_log_queue(job: BeakerJob, queue: Queue, stopped: Event):
        try:
            for job_log in beaker.job.logs(job, tail_lines=10 if tail else None, follow=True):
                if stopped.is_set():
                    return
                queue.put(job_log)
        except Exception as e:
            queue.put(e)
        finally:
            queue.put(sentinel)

    while True:
        try:
            job = get_latest_job()

            # Wait for job to be created...
            with ExitStack() as stack:
                msg = "[i]waiting on job...[/]"
                status: Status | None = None
                if not os.environ.get("GANTRY_GITHUB_TESTING"):
                    status = stack.enter_context(console.status(msg, spinner="point", speed=0.8))
                else:
                    utils.print_stdout(msg)
                if job is None:
                    while job is None:
                        if (j := get_latest_job()) is not None:
                            job = j
                        elif timeout is not None and (time.monotonic() - start_time) > timeout:
                            raise BeakerJobTimeoutError(
                                "Timed out while waiting for job to be created"
                            )
                        elif (
                            start_timeout is not None
                            and (time.monotonic() - start_time) > start_timeout
                        ):
                            raise BeakerJobTimeoutError(
                                "Timed out while waiting for job to be created"
                            )
                        else:
                            time.sleep(1.0)

                    if status is not None:
                        status.update("[i]waiting for job to launch...[/]")

                assert job is not None

                # Pull events until job is running or fails (whichever happens first)...
                events = set()
                while not (
                    beaker_utils.job_has_finalized(job)
                    or (beaker_utils.job_has_started(job) and show_logs)
                ):
                    for event in beaker.job.list_summarized_events(
                        job, sort_order=BeakerSortOrder.ascending, sort_field="latest_occurrence"
                    ):
                        event_hashable = (event.latest_occurrence.ToSeconds(), event.latest_message)
                        if event_hashable not in events:
                            events.add(event_hashable)
                            utils.print_stdout(f"✓ [i]{event.latest_message}[/]")
                            time.sleep(0.5)

                    if timeout is not None and (time.monotonic() - start_time) > timeout:
                        for callback in callbacks or []:
                            callback.on_timeout(job)
                        if beaker_utils.job_has_started(job):
                            raise BeakerJobTimeoutError(
                                f"Timed out while waiting for job '{job.id}' to finish"
                            )
                        else:
                            raise BeakerJobTimeoutError(
                                f"Timed out while waiting for job '{job.id}' to start"
                            )
                    elif (
                        start_timeout is not None
                        and not beaker_utils.job_has_started(job)
                        and (time.monotonic() - start_time) > start_timeout
                    ):
                        for callback in callbacks or []:
                            callback.on_start_timeout(job)
                        raise BeakerJobTimeoutError(
                            f"Timed out while waiting for job '{job.id}' to start"
                        )
                    else:
                        time.sleep(0.5)
                        job = beaker.job.get(job.id)

            assert job is not None

            if job.id not in on_start_called:
                for callback in callbacks or []:
                    callback.on_start(job)
                on_start_called.add(job.id)

            # Stream logs...
            if show_logs and beaker_utils.job_has_started(job):
                queue: Queue = Queue()
                stopped = Event()

                thread = Thread(target=fill_log_queue, args=(job, queue, stopped), daemon=True)
                thread.start()

                utils.print_stdout()
                rich.get_console().rule("Logs")

                last_event = time.monotonic()
                last_inactive_warning = 0.0
                while True:
                    try:
                        job_log = queue.get(timeout=1.0)
                        last_event = time.monotonic()
                        if job_log is sentinel:
                            break
                        elif isinstance(job_log, Exception):
                            raise job_log
                        else:
                            log_line = job_log.message.decode()
                            log_time = job_log.timestamp.seconds + job_log.timestamp.nanos / 1e9
                            utils.print_stdout(log_line, markup=False)
                            for callback in callbacks or []:
                                callback.on_log(job, log_line, log_time)
                    except Empty:
                        cur_time = time.monotonic()
                        if (
                            inactive_timeout is not None
                            and (cur_time - last_event) > inactive_timeout
                        ):
                            for callback in callbacks or []:
                                callback.on_inactive_timeout(job)
                            raise BeakerJobTimeoutError(
                                f"Timed out while waiting for job '{job.id}' to produce more logs"
                            )

                        if (
                            inactive_soft_timeout is not None
                            and (cur_time - last_event) > inactive_soft_timeout
                            and (cur_time - last_inactive_warning)
                            > max(inactive_soft_timeout, 3600)
                        ):
                            last_inactive_warning = cur_time
                            formatted_duration = utils.format_timedelta(
                                cur_time - last_event,
                                resolution="minutes"
                                if inactive_soft_timeout % 60 == 0
                                else "seconds",
                            )
                            log.warning(
                                f"Job appears to be inactive! No new logs within the past {formatted_duration}."
                            )
                            for callback in callbacks or []:
                                callback.on_inactive_soft_timeout(job)
                        else:
                            for callback in callbacks or []:
                                callback.on_no_new_logs(job)

                    if timeout is not None and (time.monotonic() - start_time) > timeout:
                        for callback in callbacks or []:
                            callback.on_timeout(job)
                        raise BeakerJobTimeoutError(
                            f"Timed out while waiting for job '{job.id}' to finish"
                        )

                utils.print_stdout()
                rich.get_console().rule("End logs")

            # Wait for job to finalize...
            while not beaker_utils.job_has_finalized(job):
                time.sleep(0.5)
                job = beaker.job.get(job.id)

            utils.print_stdout()

            # If job was preempted, we start over...
            if beaker_utils.job_was_preempted(job):
                utils.print_stdout(f"[yellow]Job '{job.id}' preempted.[/] ")
                preempted_job_ids.add(job.id)
                for callback in callbacks or []:
                    callback.on_preemption(job)
                job = None
                continue

            return job
        except exceptions_to_cancel_on as exc:
            utils.print_stderr(f"[red][bold]{exc.__class__.__name__}:[/] [i]{exc}[/][/]")
            beaker.workload.cancel(workload)
            utils.print_stderr("[yellow]Experiment cancelled.[/]")

            for callback in callbacks or []:
                callback.on_cancellation(job)

            if utils.is_cli_mode():
                sys.exit(1)
            else:
                raise
        except KeyboardInterrupt:
            if utils.is_interactive_terminal():
                utils.print_stderr("[yellow]Caught keyboard interrupt...[/]")
                utils.print_stderr(
                    f"You are currently following [blue u]{beaker.workload.url(workload)}[/]"
                )
                action = prompt.Prompt.ask(
                    "Press [red b]c[/] to [red]cancel[/] the workload, "
                    "[green b]r[/] to [green]resume[/] following, "
                    "or [yellow b]q[/] to [yellow]quit[/]",
                    choices=["c", "r", "q"],
                )
                if action == "c":
                    if prompt.Confirm.ask("Are you sure you'd like to cancel the experiment?"):
                        beaker.workload.cancel(workload)
                        utils.print_stderr(
                            f"[red]Experiment stopped:[/] [blue u]{beaker.workload.url(workload)}[/]"
                        )

                        for callback in callbacks or []:
                            callback.on_cancellation(job)

                        if utils.is_cli_mode():
                            sys.exit(0)
                        else:
                            raise
                elif action == "q":
                    utils.print_stdout(
                        f"See the experiment at [blue u]{beaker.workload.url(workload)}[/]"
                    )
                    utils.print_stderr(
                        f"To [yellow b]cancel[/] the workload manually, run:\n\n"
                        f"  $ gantry stop {workload.experiment.id}\n\n"
                        f"To [green b]resume following[/] the workload, run:\n\n"
                        f"  $ gantry follow --tail {workload.experiment.id}\n"
                    )
            else:
                utils.print_stderr("[yellow]Caught SIGINT...[/]")
                if auto_cancel:
                    beaker.workload.cancel(workload)
                    utils.print_stderr("[yellow]Experiment cancelled.[/]")

            if utils.is_cli_mode():
                sys.exit(1)
            else:
                raise
        finally:
            tail = True
            if stopped is not None:
                stopped.set()
                stopped = None



import binascii
import functools as ft
import hashlib
import logging
import os
import random
import tempfile
import time
from collections import defaultdict
from contextlib import ExitStack
from datetime import datetime
from fnmatch import fnmatch
from pathlib import Path
from typing import Any, Iterable, Literal, Sequence

import rich
from beaker import *
from beaker.exceptions import *
from rich import prompt

from . import constants, utils
from .aliases import PathOrStr
from .callbacks import Callback
from .exceptions import *
from .git_utils import GitRepoState
from .version import VERSION

log = logging.getLogger(__name__)


def init_client(
    workspace: str | None = None,
    yes: bool = False,
    ensure_workspace: bool = True,
    beaker_token: str | None = None,
    check_for_upgrades: bool = True,
) -> Beaker:
    Beaker.MAX_RETRIES = 10_000  # effectively retry forever
    Beaker.BACKOFF_MAX = 32

    kwargs = dict()
    if workspace is not None:
        kwargs["default_workspace"] = workspace
    if beaker_token is not None:
        kwargs["user_token"] = beaker_token
    beaker = Beaker.from_env(check_for_upgrades=check_for_upgrades, **kwargs)  # type: ignore[arg-type]

    if ensure_workspace and workspace is None:
        try:
            default_workspace = beaker.workspace.get()
            if not yes and not prompt.Confirm.ask(
                f"Using default workspace [b cyan]{default_workspace.name}[/]. [i]Is that correct?[/]"
            ):
                raise KeyboardInterrupt
        except BeakerWorkspaceNotSet:
            raise ConfigurationError(
                f"{utils.fmt_opt('--workspace')} option is required since you don't have a default workspace set"
            )
    return beaker


def is_running_in_beaker() -> bool:
    """
    Check if the current process is running inside of a Beaker job (batch or session).
    """
    # There's a number of different environment variables set by the Beaker executor.
    # Checking any one of these would suffice, but we check a couple to reduce the
    # risk of false positives.
    return "BEAKER_JOB_ID" in os.environ and "BEAKER_NODE_ID" in os.environ


def is_running_in_beaker_batch_job() -> bool:
    """
    Check if the current process is running inside a Beaker batch job (as opposed to a session).
    """
    return is_running_in_beaker() and os.environ.get("BEAKER_JOB_KIND") == "batch"


def is_running_in_gantry_batch_job() -> bool:
    """
    Check if the current process is running inside a Beaker batch job launched by Gantry.
    """
    return is_running_in_beaker_batch_job() and "GANTRY_VERSION" in os.environ


def job_was_preempted(job: BeakerJob) -> bool:
    return job.status.status == BeakerWorkloadStatus.canceled and job.status.canceled_code in (
        BeakerCancelationCode.system_preemption,
        BeakerCancelationCode.user_preemption,
    )


def job_has_started(job: BeakerJob) -> bool:
    return job.status.HasField("started")


def job_has_finalized(job: BeakerJob) -> bool:
    return job.status.HasField("finalized")


def get_latest_workload(
    beaker: Beaker,
    *,
    author_name: str | None = None,
    workspace_name: str | None = None,
    running: bool = False,
) -> BeakerWorkload | None:
    workspace = beaker.workspace.get(workspace_name)

    workloads = list(
        beaker.workload.list(
            workspace=workspace,
            finalized=not running,
            workload_type=BeakerWorkloadType.experiment,
            sort_order=BeakerSortOrder.descending,
            sort_field="created",
            author=None if author_name is None else beaker.user.get(author_name),
            limit=1,
        )
    )

    if workloads:
        return workloads[0]
    else:
        return None


def get_job(
    beaker: Beaker, wl: BeakerWorkload, task: BeakerTask | None, run: int | None = None
) -> BeakerJob | None:
    if run is None:
        return beaker.workload.get_latest_job(wl, task=task)
    else:
        # NOTE: ascending sort order on creation time is not implemented server-side yet
        jobs = list(reversed(list(beaker.job.list(task=task, sort_field="created"))))
        try:
            return jobs[run - 1]
        except IndexError:
            raise ConfigurationError(f"run number {run} does not exist")


def ensure_datasets(beaker: Beaker, *datasets: str) -> list[tuple[str, str | None, str]]:
    out = []
    for dataset_str in datasets:
        dataset_name: str
        path: str
        sub_path: str | None = None
        if dataset_str.count(":") == 1:
            dataset_name, path = dataset_str.split(":")
        elif dataset_str.count(":") == 2:
            dataset_name, sub_path, path = dataset_str.split(":")
        else:
            raise ValueError(
                f"Bad '--dataset' specification: '{dataset_str}'\n"
                f"Datasets should be in the form of 'dataset-name:/mount/location'"
                f"or 'dataset-name:sub/path:/mount/location'"
            )
        dataset_id = beaker.dataset.get(dataset_name).id
        out.append((dataset_id, sub_path, path))
    return out


def ensure_source_dataset(
    beaker: Beaker, *sources: PathOrStr, budget: str | None = None
) -> BeakerDataset:
    workspace = beaker.workspace.get()

    # Get hash of the contents of the file to create a unique name.
    sha256_hash = hashlib.sha256()
    all_source_files = []
    for source in sources:
        source = Path(source)
        if source.is_dir():
            for path in source.glob("**/*"):
                if path.is_file():
                    all_source_files.append(path)
        elif source.is_file():
            all_source_files.append(source)
        else:
            raise FileNotFoundError(source)

    for path in all_source_files:
        sha256_hash.update(str(path.resolve()).encode())
        with path.open("rb") as f:
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)

    dataset_name = f"gantry-v{VERSION}-source-{sha256_hash.hexdigest()}"

    def get_dataset() -> BeakerDataset | None:
        for dataset in beaker.dataset.list(
            workspace=workspace, name_or_description=dataset_name, results=False
        ):
            if dataset.name == dataset_name:
                return dataset
        return None

    dataset = get_dataset()
    if dataset is None:
        # Create it.
        log.info(f"Creating dataset {dataset_name} from local source files...")
        try:
            dataset = beaker.dataset.create(dataset_name, *sources, budget=budget, strip_paths=True)
            log.info(f"Dataset created: {beaker.dataset.url(dataset)}")
        except BeakerDatasetConflict:  # could be in a race with another `gantry` process.
            time.sleep(1.0)
            dataset = get_dataset()
            if dataset is None:
                raise RuntimeError(f"Failed to resolve dataset '{dataset_name}'. Please try again.")
            log.info(f"Reusing dataset {beaker.dataset.url(dataset)} for local source files")
    else:
        log.info(f"Reusing dataset {beaker.dataset.url(dataset)} for local source files")

    # Verify contents.
    ds_files = list(beaker.dataset.list_files(dataset))
    for retry in range(1, 4):
        ds_files = list(beaker.dataset.list_files(dataset))
        if len(ds_files) >= len(all_source_files):
            break
        else:
            time.sleep(1.5**retry)

    if len(ds_files) != len(all_source_files):
        raise EntrypointChecksumError(
            f"Dataset {beaker.dataset.url(dataset)} for local source files is missing the "
            f"required files. Please run again."
        )

    return dataset


def ensure_entrypoint_dataset(beaker: Beaker, budget: str | None = None) -> BeakerDataset:
    from importlib.resources import read_binary

    import gantry

    workspace = beaker.workspace.get()

    # Get hash of the local entrypoint source file.
    sha256_hash = hashlib.sha256()
    contents = read_binary(gantry, constants.ENTRYPOINT)
    sha256_hash.update(contents)

    entrypoint_dataset_name = f"gantry-v{VERSION}-{workspace.id}-{sha256_hash.hexdigest()[:6]}"

    def get_dataset() -> BeakerDataset | None:
        for dataset in beaker.dataset.list(
            workspace=workspace, name_or_description=entrypoint_dataset_name, results=False
        ):
            if dataset.name == entrypoint_dataset_name:
                return dataset
        return None

    # Ensure gantry entrypoint dataset exists.
    gantry_entrypoint_dataset = get_dataset()
    if gantry_entrypoint_dataset is None:
        # Create it.
        log.info(f"Creating entrypoint dataset {entrypoint_dataset_name}...")
        try:
            with tempfile.TemporaryDirectory() as tmpdirname:
                tmpdir = Path(tmpdirname)
                entrypoint_path = tmpdir / constants.ENTRYPOINT
                with open(entrypoint_path, "wb") as entrypoint_file:
                    entrypoint_file.write(contents)
                gantry_entrypoint_dataset = beaker.dataset.create(
                    entrypoint_dataset_name, entrypoint_path, budget=budget
                )
        except BeakerDatasetConflict:  # could be in a race with another `gantry` process.
            time.sleep(1.0)
            gantry_entrypoint_dataset = get_dataset()

    if gantry_entrypoint_dataset is None:
        raise RuntimeError(f"Failed to resolve entrypoint dataset '{entrypoint_dataset_name}'")

    # Verify contents.
    ds_files = list(beaker.dataset.list_files(gantry_entrypoint_dataset))
    for retry in range(1, 4):
        ds_files = list(beaker.dataset.list_files(gantry_entrypoint_dataset))
        if len(ds_files) >= 1:
            break
        else:
            time.sleep(1.5**retry)

    if len(ds_files) != 1:
        raise EntrypointChecksumError(
            f"Entrypoint dataset {beaker.dataset.url(gantry_entrypoint_dataset)} is missing the "
            f"required entrypoint file. Please run again."
        )

    if ds_files[0].HasField("digest"):
        digest = ds_files[0].digest
        expected_value = binascii.hexlify(digest.value).decode()
        hasher = BeakerDatasetFileAlgorithmType(digest.algorithm).hasher()
        hasher.update(contents)
        actual_value = binascii.hexlify(hasher.digest()).decode()
        if actual_value != expected_value:
            raise EntrypointChecksumError(
                f"Checksum failed for entrypoint dataset {beaker.dataset.url(gantry_entrypoint_dataset)}\n"
                f"This could be a bug, or it could mean someone has tampered with the dataset.\n"
                f"If you're sure no one has tampered with it, you can delete the dataset from "
                f"the Beaker dashboard and try again.\n"
                f"Actual digest:\n{digest}"
            )

    return gantry_entrypoint_dataset


def secret_exists(beaker: Beaker, name: str) -> bool:
    try:
        beaker.secret.get(name)
        return True
    except BeakerSecretNotFound:
        return False


def ensure_secret(beaker: Beaker, name: str, value: str) -> str:
    if not value:
        raise InvalidSecretError(f"Value for secret '{name}' can't be empty")

    # Create a unique name for this secret based on the env var name and a hash
    # of the value.
    sha256_hash = hashlib.sha256()
    sha256_hash.update(value.encode(errors="ignore"))
    secret = f"{name}_{sha256_hash.hexdigest()[:8]}"
    attempts = 1
    while True:
        try:
            s = beaker.secret.get(secret)
        except BeakerSecretNotFound:
            beaker.secret.write(secret, value)
            break

        if beaker.secret.read(s) == value:
            break

        # It's highly unlikely to get a naming conflict here but we handle it anyway.
        secret = f"{name}_{sha256_hash.hexdigest()[:8]}_{attempts}"
        attempts += 1

    return secret


def resolve_group(
    beaker: Beaker,
    group_name: str,
    workspace_name: str | None = None,
    fall_back_to_default_workspace: bool = True,
) -> BeakerGroup | None:
    workspace: BeakerWorkspace | None = None
    if workspace_name is not None or fall_back_to_default_workspace:
        workspace = beaker.workspace.get(workspace_name)

    group_owner: str | None = None
    if "/" in group_name:
        group_owner, group_name = group_name.split("/", 1)

    for group in beaker.group.list(workspace=workspace, name_or_description=group_name):
        if group_owner is not None:
            if f"{group_owner}/{group_name}" == group.full_name:
                return group
        elif group_name == group.name:
            return group
    return None


def get_job_status_str(job: BeakerJob):
    status = job.status.status
    canceled_code = job.status.canceled_code
    if status == BeakerWorkloadStatus.canceled:
        if canceled_code == BeakerCancelationCode.sibling_task_failed:
            return "canceled due to sibling task failure"
        else:
            return "canceled"
    elif status == BeakerWorkloadStatus.failed:
        if job.status.HasField("exit_code") and job.status.exit_code != 0:
            return f"failed with exit code {job.status.exit_code}"
        else:
            return "failed"
    else:
        return str(BeakerWorkloadStatus(status).name)


def show_all_jobs(beaker: Beaker, workload: BeakerWorkload):
    utils.print_stdout("Tasks:")
    task_name: str | None = None
    for task in workload.experiment.tasks:
        task_name = task.name
        job = beaker.workload.get_latest_job(workload, task=task)
        assert job is not None
        status_str = get_job_status_str(job)
        style = "[white]"
        if job.status.status == BeakerWorkloadStatus.failed:
            style = "[red]"
        elif job.status.status == BeakerWorkloadStatus.canceled:
            style = "[yellow]"
        utils.print_stdout(
            f"❯ {style}'{task_name}'[/] {status_str} - see [cyan u]{beaker.job.url(job)}[/]"
        )

    assert task_name is not None
    utils.print_stdout(
        f"\nYou can show the logs for a particular task by running:\n"
        f"[i][blue]gantry[/] [cyan]logs {workload.experiment.id} --tail=1000 --task={task_name}[/][/]"
    )


def group_url(beaker: Beaker, group: BeakerGroup) -> str:
    # NOTE: work-around for https://github.com/allenai/beaker-web/issues/1109, short group URLs
    # don't resolve at the moment.
    org_name = beaker.org_name
    workspace_name = beaker.workspace.get(group.workspace_id).name.split("/", 1)[-1]
    return f"{beaker.config.agent_address}/orgs/{org_name}/workspaces/{workspace_name}/groups/{group.id}"


def get_gpu_type(beaker: Beaker, cluster: BeakerCluster) -> str | None:
    nodes = list(beaker.node.list(cluster=cluster, limit=1))
    if nodes:
        try:
            return BeakerGpuType(nodes[0].node_resources.gpu_type).name.replace("_", " ")
        except ValueError:
            return None
    else:
        return None


def filter_clusters_by_name(
    beaker: Beaker, clusters: Iterable[BeakerCluster], patterns: Sequence[str]
) -> list[BeakerCluster]:
    matches = set()
    matches_by_pattern: dict[str, int] = defaultdict(int)
    final_clusters = []
    for cl in clusters:
        cl_aliases = list(cl.aliases) + [cl.name]
        for pattern in patterns:
            og_pattern = pattern
            if "/" in pattern:
                org, pattern = pattern.split("/", 1)
            else:
                org = beaker.org_name

            if cl.organization_name != org:
                continue

            if "*" in pattern:
                if not any([fnmatch(alias, pattern) for alias in cl_aliases]):
                    continue
            elif not any([alias == pattern for alias in cl_aliases]):
                continue

            matches_by_pattern[og_pattern] += 1
            if cl.id not in matches:
                matches.add(cl.id)
                final_clusters.append(cl)

    for pattern in patterns:
        if matches_by_pattern[pattern] == 0:
            raise ConfigurationError(f"'{pattern}' didn't match any allowed clusters")

    return final_clusters


def filter_clusters_by_tags(
    beaker: Beaker,
    clusters: Iterable[BeakerCluster],
    tags: Sequence[str],
) -> list[BeakerCluster]:
    del beaker
    final_clusters = []
    for cl in clusters:
        cl_tags = set(cl.tags)
        if all([tag in cl_tags for tag in tags]):
            final_clusters.append(cl)
    return final_clusters


def filter_clusters_by_gpu_type(
    beaker: Beaker,
    clusters: Iterable[BeakerCluster],
    gpu_types: Sequence[str],
) -> list[BeakerCluster]:
    final_clusters = []
    for cl in clusters:
        cl_gpu_type = get_gpu_type(beaker, cl)
        if not cl_gpu_type:
            continue
        for pattern in gpu_types:
            if pattern.lower() in cl_gpu_type.lower():
                final_clusters.append(cl)
                break
    return final_clusters


def download_logs(
    beaker: Beaker,
    job: BeakerJob,
    head_lines: int | None = None,
    tail_lines: int | None = None,
    since: datetime | None = None,
    follow: bool = True,
    out_path: PathOrStr | None = None,
) -> tuple[BeakerJob, int]:
    if out_path is not None:
        out_path = Path(out_path)
        out_path.parent.mkdir(parents=True, exist_ok=True)
    else:
        utils.print_stdout()
        rich.get_console().rule("Logs")

    if tail_lines is not None and head_lines is not None:
        raise ConfigurationError("can't use both 'head_lines' and 'tail_lines' options together")
    if tail_lines is not None and tail_lines < 0:
        raise ConfigurationError("'tail_lines' must be non-negative")
    if head_lines is not None and head_lines < 0:
        raise ConfigurationError("'head_lines' must be non-negative")

    n_lines = 0
    with ExitStack() as stack:
        out_file = None
        if out_path is not None:
            out_file = stack.enter_context(open(out_path, "wb"))

        for job_log in beaker.job.logs(job, follow=follow, tail_lines=tail_lines, since=since):
            n_lines += 1
            if out_file is None:
                utils.print_stdout(job_log.message.decode(), markup=False)
            else:
                out_file.write(job_log.message)
                out_file.write(b"\n")
            if head_lines is not None and n_lines >= head_lines:
                break

    if out_path is None:
        utils.print_stdout()
        rich.get_console().rule("End logs")

    return beaker.job.get(job.id), n_lines


def display_results(
    beaker: Beaker,
    workload: BeakerWorkload,
    job: BeakerJob,
    info_header: str | None = None,
    callbacks: Sequence[Callback] | None = None,
):
    status = job.status.status
    runtime = job.status.exited - job.status.started  # type: ignore
    results_ds = beaker.dataset.get(job.assignment_details.result_dataset_id)

    if status == BeakerWorkloadStatus.succeeded:
        utils.print_stdout(
            f"[b green]\N{check mark}[/] [b cyan]{beaker.user_name}/{workload.experiment.name}[/] ({workload.experiment.id}) completed successfully.\n"
        )

    if info_header:
        utils.print_stdout(info_header)

    utils.print_stdout(
        f"[b]Results:[/] [blue u]{beaker.dataset.url(results_ds)}[/]\n"
        f"[b]Runtime:[/] {utils.format_timedelta(runtime)}"
    )

    metrics: dict[str, Any] | None = None
    if job.metrics:
        from google.protobuf.json_format import MessageToDict

        metrics = MessageToDict(job.metrics)
        utils.print_stdout("[b]Metrics:[/]", metrics, highlight=True)

    if status in (BeakerWorkloadStatus.canceled, BeakerWorkloadStatus.failed):
        print()
        if len(list(workload.experiment.tasks)) > 1:
            show_all_jobs(beaker, workload)
            utils.print_stdout()

        for callback in callbacks or []:
            callback.on_failure(job, metrics=metrics, results_ds=results_ds)

        raise ExperimentFailedError(
            f"Job {get_job_status_str(job)}, see {beaker.workload.url(workload)} for details"
        )
    elif status == BeakerWorkloadStatus.succeeded:
        for callback in callbacks or []:
            callback.on_success(job, metrics=metrics, results_ds=results_ds)
    else:
        raise ValueError(f"unexpected workload status '{status}'")


def build_experiment_spec(
    beaker: Beaker,
    *,
    task_name: str,
    clusters: list[str],
    task_resources: BeakerTaskResources,
    arguments: list[str],
    entrypoint_dataset: str,
    git_repo: GitRepoState,
    budget: str | None = None,
    group_names: list[str] | None = None,
    description: str | None = None,
    beaker_image: str | None = None,
    docker_image: str | None = None,
    gh_token_secret: str | None = constants.GITHUB_TOKEN_SECRET,
    conda_file: PathOrStr | None = None,
    conda_env: str | None = None,
    python_manager: Literal["uv", "conda"] | None = None,
    system_python: bool = False,
    uv_venv: str | None = None,
    uv_extras: Sequence[str] | None = None,
    uv_all_extras: bool | None = None,
    uv_torch_backend: str | None = None,
    datasets: list[tuple[str, str | None, str]] | None = None,
    env: list[tuple[str, str]] | None = None,
    env_secrets: list[tuple[str, str]] | None = None,
    dataset_secrets: list[tuple[str, str]] | None = None,
    priority: str | BeakerJobPriority | None = None,
    install: str | None = None,
    no_python: bool = False,
    replicas: int | None = None,
    leader_selection: bool = False,
    host_networking: bool = False,
    propagate_failure: bool | None = None,
    propagate_preemption: bool | None = None,
    synchronized_start_timeout: str | None = None,
    task_timeout: str | None = None,
    mounts: list[tuple[str, str]] | None = None,
    weka_buckets: list[tuple[str, str]] | None = None,
    uploads: list[tuple[str, str]] | None = None,
    hostnames: list[str] | None = None,
    preemptible: bool | None = None,
    retries: int | None = None,
    results: str = constants.RESULTS_DIR,
    runtime_dir: str = constants.RUNTIME_DIR,
    exec_method: Literal["exec", "bash"] = "exec",
    torchrun: bool = False,
    skip_nccl_setup: bool = False,
    default_python_version: str = utils.get_local_python_version(),
    pre_setup: str | None = None,
    post_setup: str | None = None,
) -> BeakerExperimentSpec:
    if exec_method not in ("exec", "bash"):
        raise ConfigurationError(
            f"expected one of 'exec' or 'bash' for {utils.fmt_opt('--exec-method')}, but got '{exec_method}'."
        )

    task_spec = (
        BeakerTaskSpec.new(
            task_name,
            beaker_image=beaker_image,
            docker_image=docker_image,
            result_path=results,
            command=["bash", "/gantry/entrypoint.sh"],
            arguments=arguments,
            resources=task_resources,
            priority=priority,
            preemptible=preemptible,
            replicas=replicas,
            leader_selection=leader_selection,
            host_networking=host_networking,
            propagate_failure=propagate_failure,
            propagate_preemption=propagate_preemption,
            synchronized_start_timeout=synchronized_start_timeout,
            timeout=task_timeout,
        )
        .with_env_var(name="GANTRY_VERSION", value=VERSION)
        .with_env_var(name="GITHUB_REPO", value=git_repo.repo)
        .with_env_var(name="GIT_REF", value=git_repo.ref)
        .with_env_var(name="GANTRY_TASK_NAME", value=task_name)
        .with_env_var(name="RESULTS_DIR", value=results)
        .with_env_var(name="GANTRY_RUNTIME_DIR", value=runtime_dir)
        .with_env_var(name="GANTRY_EXEC_METHOD", value=exec_method)
        .with_dataset("/gantry", beaker=entrypoint_dataset)
    )

    if torchrun:
        task_spec = task_spec.with_env_var(name="GANTRY_USE_TORCHRUN", value="1")
        if replicas and leader_selection:
            task_spec = task_spec.with_env_var(
                name="GANTRY_RDZV_ID", value=str(random.randint(0, 999))
            )
            task_spec = task_spec.with_env_var(
                name="GANTRY_RDZV_PORT", value=str(random.randint(29_000, 29_999))
            )

    if git_repo.branch is not None:
        task_spec = task_spec.with_env_var(name="GIT_BRANCH", value=git_repo.branch)

    if clusters:
        task_spec = task_spec.with_constraint(cluster=clusters)

    if hostnames:
        task_spec = task_spec.with_constraint(hostname=hostnames)

    if gh_token_secret is not None:
        task_spec = task_spec.with_env_var(name="GITHUB_TOKEN", secret=gh_token_secret)

    if skip_nccl_setup:
        task_spec = task_spec.with_env_var(name="GANTRY_SKIP_NCCL_SETUP", value="1")

    if no_python:
        task_spec = task_spec.with_env_var(name="GANTRY_NO_PYTHON", value="1")

        if (
            python_manager is not None
            or system_python
            or uv_venv is not None
            or uv_all_extras is not None
            or uv_extras
            or uv_torch_backend is not None
            or conda_env is not None
            or conda_file is not None
        ):
            raise ConfigurationError(
                f"other python options can't be used with {utils.fmt_opt('--no-python')}"
            )
    else:

        @ft.cache
        def has_project_file() -> bool:
            return (
                git_repo.is_in_tree("pyproject.toml")
                or git_repo.is_in_tree("setup.py")
                or git_repo.is_in_tree("setup.cfg")
            )

        task_spec = task_spec.with_env_var(
            name="GANTRY_DEFAULT_PYTHON_VERSION",
            value=default_python_version,
        )

        if system_python:
            task_spec = task_spec.with_env_var(
                name="GANTRY_USE_SYSTEM_PYTHON",
                value="1",
            )

        if python_manager is None:
            if (
                conda_env is not None
                or conda_file is not None
                or git_repo.is_in_tree("environment.yml")
                or git_repo.is_in_tree("environment.yaml")
            ):
                python_manager = "conda"
            else:
                python_manager = "uv"
        elif python_manager not in {"uv", "conda"}:
            raise ConfigurationError(
                f"unknown option for {utils.fmt_opt('--python-manager')}: '{python_manager}'. Should be either 'uv' or 'conda'."
            )

        task_spec = task_spec.with_env_var(
            name="GANTRY_PYTHON_MANAGER",
            value=python_manager,
        )

        if python_manager == "uv":
            if conda_env is not None or conda_file is not None:
                raise ConfigurationError(
                    f"{utils.fmt_opt('--conda-*')} options are only relevant when using the conda python manager "
                    f"({utils.fmt_opt('--python-manager')}=conda)."
                )

            if uv_venv is not None:
                if system_python:
                    raise ConfigurationError(
                        f"{utils.fmt_opt('--system-python')} flag is incompatible with "
                        f"{utils.fmt_opt('--uv-venv')} option."
                    )

                task_spec = task_spec.with_env_var(
                    name="GANTRY_UV_VENV",
                    value=uv_venv,
                )

            if uv_all_extras is None:
                if not uv_extras and has_project_file():
                    uv_all_extras = True
                else:
                    uv_all_extras = False
            elif uv_extras:
                raise ConfigurationError(
                    f"{utils.fmt_opt('--uv-all-extras/--uv-no-extras')} is mutually exclusive "
                    f"with {utils.fmt_opt('--uv-extra')}."
                )

            if uv_all_extras:
                if not has_project_file():
                    raise ConfigurationError(
                        f"{utils.fmt_opt('--uv-all-extras')} is only valid when you have a pyproject.toml, setup.py, or setup.cfg file."
                    )

                task_spec = task_spec.with_env_var(name="GANTRY_UV_ALL_EXTRAS", value="1")

            if uv_extras:
                if not has_project_file():
                    raise ConfigurationError(
                        f"{utils.fmt_opt('--uv-extra')} is only valid when you have a pyproject.toml, setup.py, or setup.cfg file."
                    )

                task_spec = task_spec.with_env_var(
                    name="GANTRY_UV_EXTRAS", value=" ".join(uv_extras)
                )

            if uv_torch_backend is not None:
                task_spec = task_spec.with_env_var(name="UV_TORCH_BACKEND", value=uv_torch_backend)
        elif python_manager == "conda":
            if (
                uv_venv is not None
                or uv_extras
                or uv_all_extras is not None
                or uv_torch_backend is not None
            ):
                raise ConfigurationError(
                    f"{utils.fmt_opt('--uv-*')} options are only relevant when using the uv python "
                    f"manager ({utils.fmt_opt('--python-manager')}=uv)."
                )

            if conda_env is not None:
                if system_python:
                    raise ConfigurationError(
                        f"{utils.fmt_opt('--system-python')} flag is incompatible with "
                        f"{utils.fmt_opt('--conda-env')} option."
                    )

                task_spec = task_spec.with_env_var(
                    name="GANTRY_CONDA_ENV",
                    value=conda_env,
                )

            if conda_file is not None:
                task_spec = task_spec.with_env_var(
                    name="GANTRY_CONDA_FILE",
                    value=str(conda_file),
                )
            else:
                for path in ("environment.yml", "environment.yaml"):
                    if git_repo.is_in_tree(path):
                        task_spec = task_spec.with_env_var(
                            name="GANTRY_CONDA_FILE",
                            value=path,
                        )
                        break

    if install is not None:
        task_spec = task_spec.with_env_var(name="GANTRY_INSTALL_CMD", value=install)

    if pre_setup is not None:
        task_spec = task_spec.with_env_var(name="GANTRY_PRE_SETUP_CMD", value=pre_setup)

    if post_setup is not None:
        task_spec = task_spec.with_env_var(name="GANTRY_POST_SETUP_CMD", value=post_setup)

    if datasets:
        for dataset_id, sub_path, path in datasets:
            task_spec = task_spec.with_dataset(path, beaker=dataset_id, sub_path=sub_path)

    for secret, mount_path in dataset_secrets or []:
        task_spec = task_spec.with_dataset(mount_path, secret=secret)

    if mounts:
        for source, target in mounts:
            task_spec = task_spec.with_dataset(target, host_path=source)

    if weka_buckets:
        for source, target in weka_buckets:
            task_spec = task_spec.with_dataset(target, weka=source)

    if uploads:
        for source, target in uploads:
            if Path(target).suffix:
                raise ConfigurationError(
                    f"The target mount path for uploads must be a directory, not a filename. Got: '{target}'"
                )
            source_dataset = ensure_source_dataset(beaker, source, budget=budget)
            task_spec = task_spec.with_dataset(target, beaker=source_dataset.id)

    for name, val in env or []:
        task_spec = task_spec.with_env_var(name=name, value=val)

    for name, secret in env_secrets or []:
        task_spec = task_spec.with_env_var(name=name, secret=secret)

    return BeakerExperimentSpec(
        description=description,
        budget=budget,
        tasks=[task_spec],
        groups=group_names,
        retry=None if not retries else BeakerRetrySpec(allowed_task_retries=retries),
    )



from __future__ import annotations

import dataclasses
import json
import logging
import re
import subprocess
import warnings
from dataclasses import dataclass
from functools import cache, cached_property
from pathlib import Path

import requests
from git import InvalidGitRepositoryError
from git.refs import Head
from git.repo import Repo

from . import utils
from .aliases import PathOrStr
from .exceptions import *

__all__ = ["GitRepoState"]

log = logging.getLogger(__name__)

_GITHUB_NAME_RE = re.compile(r"^[a-zA-Z0-9._-]+$")


def _parse_git_remote_url(url: str) -> tuple[str, str]:
    """
    Parse a git remote URL into a GitHub (account, repo) pair.

    :raises InvalidRemoteError: If the URL can't be parsed correctly.
    """
    if "github.com" not in url:
        if url.count("/") == 1:
            account, repo = url.split("/")
            if _GITHUB_NAME_RE.match(account) and _GITHUB_NAME_RE.match(repo):
                return account, repo

        raise InvalidRemoteError(f"Remote ('{url}') must point to a GitHub repo")

    try:
        account, repo = url.split("github.com", 1)[-1].strip("/:").split(".git")[0].split("/")
    except ValueError:
        raise InvalidRemoteError(f"Failed to parse GitHub repo path from remote '{url}'")

    return account, repo


@cache
def _resolve_repo() -> Repo:
    try:
        return Repo(".")
    except InvalidGitRepositoryError as e:
        raise GitError(
            f"gantry must be run from the ROOT of a valid git repository "
            f"unless {utils.fmt_opt('--remote')} is provided to use a remote repository."
        ) from e


@dataclass
class GitRepoState:
    """
    Represents the state of a local git repository.

    .. tip::
        Use :meth:`from_env()` to instantiate this class.
    """

    repo: str
    """
    The repository name, e.g. ``"allenai/beaker-gantry"``.
    """
    repo_url: str
    """
    The repository URL for cloning, e.g. ``"https://github.com/allenai/beaker-gantry"``.
    """
    ref: str
    """
    The current commit ref/SHA.
    """
    branch: str | None = None
    """
    The current active branch, if any.
    """
    _is_remote: bool = dataclasses.field(repr=False, default=False)

    @property
    def is_dirty(self) -> bool:
        """
        If the local repository state is dirty (uncommitted changes).
        """
        if self._is_remote:
            return False
        repo = _resolve_repo()
        return repo.is_dirty()

    @cached_property
    def is_public(self) -> bool:
        """
        If the repository is public.
        """
        response = requests.get(self.repo_url)
        if response.status_code not in {200, 404}:
            response.raise_for_status()
        return response.status_code == 200

    @property
    def short_ref(self) -> str:
        """
        Short, 7-character version of the current :data:`ref`.
        """
        if len(self.ref) == 40 and self.ref.isalnum():
            return self.ref[:7]
        else:
            return self.ref

    @property
    def ref_url(self) -> str:
        """
        The URL to the current :data:`ref`.
        """
        return f"{self.repo_url}/commit/{self.short_ref}"

    @property
    def branch_url(self) -> str | None:
        """
        The URL to the current active :data:`branch`.
        """
        if self.branch is None:
            return None
        else:
            return f"{self.repo_url}/tree/{self.branch}"

    @cached_property
    def commit_message(self) -> str | None:
        """
        Full commit message.
        """
        if self._is_remote:
            return None
        repo = _resolve_repo()
        try:
            return str(repo.commit(self.ref).message)
        except Exception:
            return None

    def short_commit_message(self, max_length: int = 50) -> str | None:
        """
        The commit message, truncated to ``max_length`` characters.
        """
        if self.commit_message is None:
            return None
        msg = self.commit_message.split("\n")[0].strip()
        if len(msg) <= max_length:
            return msg
        else:
            return msg[: max_length - 1] + "…"

    def is_in_tree(self, path: PathOrStr) -> bool:
        """
        Check if a file is in the tree.
        """
        if not self._is_remote:
            path = Path(path).resolve().relative_to(Path("./").resolve())
            repo = _resolve_repo()
            tree = repo.commit(self.ref).tree
            return str(path) in tree

        try:
            res = subprocess.run(
                [
                    "gh",
                    "api",
                    f"repos/{self.repo}/contents/{path}?ref={self.ref}",
                    "--jq",
                    ".name",
                ],
                capture_output=True,
                text=True,
            )
        except FileNotFoundError:
            raise ConfigurationError(
                f"Attempted to use the GitHub CLI to pull metadata about the remote repo '{self.repo}', "
                f"however it appears that it's not installed. "
                f"Please ensure the GitHub CLI is installed and try again."
            )

        try:
            res.check_returncode()
            return True
        except Exception:
            try:
                status = int(json.loads(res.stdout)["status"])
                if status == 404:
                    return False
            except Exception:
                pass

            raise

    @classmethod
    def from_remote(
        cls, remote_url: str, ref: str | None = None, branch: str | None = None
    ) -> GitRepoState:
        """
        Instantiate this class from a remote repository.
        """
        account, repo_name = _parse_git_remote_url(remote_url)
        if ref is None:
            if branch is None:
                raise ConfigurationError(
                    f"Either {utils.fmt_opt('--ref')} or {utils.fmt_opt('--branch')} is required "
                    f"when using remote repositories."
                )

            try:
                res = subprocess.run(
                    [
                        "gh",
                        "api",
                        f"repos/{account}/{repo_name}/commits",
                        "-f",
                        f"sha={branch}",
                        "--method",
                        "GET",
                        "--jq",
                        ".[0].sha",
                    ],
                    capture_output=True,
                    text=True,
                )
            except FileNotFoundError:
                raise ConfigurationError(
                    f"Since {utils.fmt_opt('--ref')} was not provided, attempted to determine the SHA of the latest "
                    f"commit automatically using the GitHub CLI, however it appears that it's not installed. "
                    f"Please provide a {utils.fmt_opt('--ref')} (SHA of a commit to use) or ensure the GitHub CLI is installed."
                )

            try:
                res.check_returncode()
            except Exception:
                raise ConfigurationError(
                    f"Since {utils.fmt_opt('--ref')} was not provided, attempted to determine the SHA of the latest "
                    f"commit automatically using the GitHub CLI, however this failed with:\n"
                    f"{res.stderr}\n"
                    f"You can avoid this issue by providing a {utils.fmt_opt('--ref')} (SHA of a commit to use)."
                )

            ref = res.stdout.strip()

        return cls(
            repo=f"{account}/{repo_name}",
            repo_url=f"https://github.com/{account}/{repo_name}",
            ref=ref,
            branch=branch,
            _is_remote=True,
        )

    @classmethod
    def from_env(cls, ref: str | None = None, branch: str | None = None) -> GitRepoState:
        """
        Instantiate this class from the root of a git repository.

        :raises ~gantry.exceptions.GitError: If this method isn't called from the
            root of a valid git repository.
        :raises ~gantry.exceptions.UnpushedChangesError: If there are unpushed commits.
        """
        from .beaker_utils import is_running_in_gantry_batch_job

        repo = _resolve_repo()
        git_ref = ref or str(repo.commit())
        remote = repo.remote()
        account, repo_name = _parse_git_remote_url(remote.url)

        # Check if the ref exists on the remote (if not, it's likely that there are unpushed commits).
        try:
            res = subprocess.run(
                [
                    "gh",
                    "api",
                    f"repos/{account}/{repo_name}/commits/{git_ref}",
                    "--jq={sha: .sha, status: .status}",
                ],
                capture_output=True,
                text=True,
            )
        except FileNotFoundError:
            msg = (
                "Attempted to use the GitHub CLI to validate that the commit exists on the remote, "
                "however it appears that it's not installed. "
                "Please ensure the GitHub CLI is installed to avoid this warning message."
            )
            if utils.is_cli_mode():
                utils.print_stderr(f"[yellow]{msg}[/]")
            elif not is_running_in_gantry_batch_job():
                warnings.warn(msg, UserWarning)
        else:
            try:
                output = json.loads(res.stdout)
            except json.JSONDecodeError:
                output = {}

            if not (output.get("sha") or "").startswith(git_ref):  # in case the ref is abbreviated
                if str(output.get("status")) in {"404", "422"}:
                    raise UnpushedChangesError(
                        f"Current git ref '{git_ref}' does not appear to exist on the remote!\n"
                        "Please push your changes and try again."
                    )
                else:
                    msg = (
                        f"Unexpected response from the GitHub API while validating git ref '{git_ref}' on remote:\n"
                        f"{res.stdout}\n{res.stderr}"
                    )
                    warnings.warn(msg, RuntimeWarning)

        # Resolve branch.
        branch_name: str | None = branch
        if branch_name is None:
            active_branch: Head | None = None
            try:
                active_branch = repo.active_branch
            except TypeError:
                pass
            else:
                remote_branch = active_branch.tracking_branch()
                if remote_branch is not None:
                    assert remote_branch.name.startswith(remote_branch.remote_name + "/")
                    remote = repo.remote(remote_branch.remote_name)
                    branch_name = remote_branch.name.replace(remote_branch.remote_name + "/", "", 1)

        return cls(
            repo=f"{account}/{repo_name}",
            repo_url=f"https://github.com/{account}/{repo_name}",
            ref=git_ref,
            branch=branch_name,
        )



import os

from .version import VERSION

DEFAULT_IMAGE = "beaker-py/gantry"
VERSIONED_DEFAULT_IMAGE = f"beaker-py/gantry-v{VERSION}"
ENTRYPOINT = "entrypoint.sh"
GITHUB_TOKEN_SECRET = "GITHUB_TOKEN"
RUNTIME_DIR = os.environ.get("GANTRY_RUNTIME_DIR", "/gantry-runtime")
RESULTS_DIR = os.environ.get("RESULTS_DIR", "/results")
METRICS_FILE = f"{RESULTS_DIR}/metrics.json"
